/*
  SERWER sieci wymiany plikow wg. protokolu grupy I4, z srody 8:00
  Jacek Pospychaa (66301)
 */

#include <unistd.h>
#include <signal.h>
#include <fcntl.h>

#include <sys/stat.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/sem.h>
#include <sys/wait.h>

#include <strings.h>
#include <string.h>
#include <stdarg.h>
#include <time.h>
#include <stdio.h>
#include <dirent.h>
#include <errno.h>

#include "66301_EXTENDEDprotocol.h"
#include "66301_common.h"

#define H66301_SERVER

#define RUNNING_DIR	"/tmp"
#define LOCK_FILE	RUNNING_DIR "/66301_server.lock"
#define LOG_FILE	RUNNING_DIR "/66301_server.log"

#define MAX_CLIENTS_DEFAULT             100
#define MAX_FILES_PERCLIENT_DEFAULT     200
#define MAX_DOWNBYTES_PERCLIENT_DEFAULT 1000000

/* Nazwy wykorzystywanych semaforw. 
     USERSCOUNT_SEM to liczba dostpnych uytkownikw do zalogowania
     SERV_MEM_SEM zapewnia wyczno przy zapisywaniu do struktury serv_mem zmiennych itp. 
 */
// #define USERSCOUNT_SEM 0
#define SERV_MEM_SEM 1

/* staa definiujca, jakie CLIENTID maj wiadomoci przychodzce do serwera
   od nowych klientw, ktrzy chc si zalogowa 
 */
#define CLIENTID_LOGIN_REQUEST 1

#define SERVER_DIR "66301_server_data"
#define DATA_DIR SERVER_DIR "/files"
#define PERM_DIR SERVER_DIR "/perms"

/* shared memory structure
 */
typedef struct user_info {
  char name[PROTO_ARGSIZE+1];
  pid_t childpid;
  int clientid;
  int filescount;
  int transfer;
  int admin;
} user_info;

#define USERS_INFO_SIZE 20

typedef struct users_info_mem {
  int shmid;
  user_info users[USERS_INFO_SIZE];
  struct users_info_mem* next_users;
} users_info_mem;

typedef struct serv_mem {
  int msgid;
  int shmid;
  int semid;
  int servpid;
  int config_vars[3];
  int filescount;
  int transfer;
  int clients;
  int leaving;
  users_info_mem* users;
} serv_mem;

/* indeksy tablicy config_vars w serv_mem: */
#define MAX_CLIENTS             0
#define MAX_FILES_PERCLIENT     1
#define MAX_DOWNBYTES_PERCLIENT 2

void remove_shared_space(void);
void close_sessions(void);
void unregister_connection(user_info* mem);
void safe_exit(int signal);
void termination_in_progress(int signal);
void log_message(char* format, ...);

void p(int semnum);
void v(int semnum);

serv_mem* shared_space_location(serv_mem* pointer);
user_info* user_info_location(user_info* pointer);

extern int errno;

/**************************************
 * signal handling
 **************************************/

/* bezpieczne wyjcie
 */
void safe_exit(int signal)
{
  DEBUGMSG(TRACE_DBG, ("safe_exit\n"));
  //  DEBUGMSG(GENERAL_DBG, ("safe exit...\n"));
  serv_mem* mem = shared_space_location(NULL);

  if (mem == NULL)
    {
      DEBUGMSG(GENERAL_DBG, ("safe_exit: shared mem not set. It's critical\n"));
      kill(0, SIGKILL);
    }
  else
    {
      if (mem->servpid == getppid()) // child
	{
	  DEBUGMSG(GENERAL_DBG, ("child process %d leaving due to %d signal\n", getpid(), signal));
	  unregister_connection(user_info_location(NULL));
	  exit(0);
	}
      else if (mem->servpid == getpid()) // main
	{
	  if (! mem->leaving)
	    {
	      DEBUGMSG(GENERAL_DBG, ("server main process leaving due to %d signal\n", signal));
	      log_message("leaving on %d signal", signal);
	      mem->leaving=1;
	      close_sessions();
	      close_queue(mem->msgid);
	      while (wait(NULL)>0); // poczekajmy na wszystkie dzieci
	      remove_shared_space();
	      unlink(LOCK_FILE);
	      exit(0);
	    }
	}
      else // nie child, nie main, moe sierota?
	{
	  DEBUGMSG(GENERAL_DBG, ("leaving on %d signal\n", signal));
	  exit(0);
	}
    }
}

/**************************************
 * message output
 **************************************/
void log_message(char* format, ...)
{
  va_list ap;
  char timestr[24];
  time_t now = time(NULL);
  FILE* logfile;
  logfile = fopen(LOG_FILE, "a");
  if (!logfile) return;

  strftime(timestr, 24, "%T", localtime(&now));

  va_start(ap, format);
  fprintf(logfile, "[%s] 66301_server (pid %d): ", timestr, getpid());
  vfprintf(logfile, format, ap);
  fprintf(logfile, "\n");
  va_end(ap);
  fclose(logfile);
}

/**************************************
 * directory functions
 **************************************/

/*
 Usuwa rekurencyjnie katalog/plik ze wszystkim co jest w rodku
 */
void rmr(char *dirname)
{
  DEBUGMSG(TRACE_DBG, ("rmr\n"));
  DIR* d;
  struct dirent *d_ent;
  char ent_name[2*NAME_MAX + 1] = {0};

  if ((d = opendir(dirname)) == NULL)
    {
      if (unlink(dirname) == -1)
	log_message("unlink");
    }
  else
    {
      while ((d_ent = readdir(d)) != NULL)
	{
	  if ((strlen(d_ent->d_name)==1 && strcmp(".", d_ent->d_name)==0) || 
	      (strlen(d_ent->d_name)==2 && strcmp("..", d_ent->d_name)==0))
	    continue;

	  snprintf(ent_name, 2*NAME_MAX + 1, "%s/%s", dirname, d_ent->d_name);
	  rmr(ent_name);
	}
      
      closedir(d);
      rmdir(dirname);
    }
}

void ls(int userid, char *username, char **dest, int *destsize)
{
  DEBUGMSG(TRACE_DBG, ("ls\n"));
  DIR* d;
  struct dirent *d_ent;
  int used;
  char dirname[strlen(DATA_DIR) + 10] = {0};

  snprintf(dirname, strlen(DATA_DIR) + 9, DATA_DIR "/%d", userid);

  if ((d = opendir(dirname)) != NULL)
    {
      while ((d_ent = readdir(d)) != NULL)
	{
	  if ((strlen(d_ent->d_name)==1 && strcmp(".", d_ent->d_name)==0) || 
	      (strlen(d_ent->d_name)==2 && strcmp("..", d_ent->d_name)==0))
	    continue;

	  used = snprintf(*dest, *destsize, "%s %s\n", username, d_ent->d_name);
	  *dest = (char*)((unsigned int)*dest + (unsigned int)used);
	  *destsize -= used;
	}
      closedir(d);
    }
}

/* Sprawdza, czy uzytkownik ma katalog domowy (return 1), czy nie (return 0)
 */
int user_has_dir(int clientid)
{
  DEBUGMSG(TRACE_DBG, ("user_has_dir\n"));
  char ent_name[2*NAME_MAX + 1] = {0};
  DIR* d;

  snprintf(ent_name, 2*NAME_MAX + 1, DATA_DIR "/%d", clientid);
  if ((d = opendir(ent_name)) == NULL)
    {
      DEBUGMSG(TRACE_DBG, ("dir %s not exist\n", ent_name));
      return 0;
    }
  closedir(d);
  DEBUGMSG(TRACE_DBG, ("dir %s exists\n", ent_name));
  return 1;
}

/* Tworzy katalog uytkownika
 */
int mk_user_dir(int clientid)
{
  DEBUGMSG(TRACE_DBG, ("mk_user_dir\n"));
  char ent_name[2*NAME_MAX + 2] = {0};
  char perment_name[2*NAME_MAX + 2] = {0};

  if (user_has_dir(clientid)) 
    {
      return 0;
    }

  snprintf(ent_name, 2*NAME_MAX + 1, DATA_DIR "/%d", clientid);
  snprintf(perment_name, 2*NAME_MAX + 1, PERM_DIR "/%d", clientid);

  if ((mkdir(ent_name, 0755) == -1) || (mkdir(perment_name, 0755) == -1))
    {
      log_message("mkuserdir");
      return -1;
    }

  DEBUGMSG(TRACE_DBG, ("dir %s created\n", ent_name));
  return 0;
}

/* Usuwa katalog uytkownika
 */
int rm_user_dir(int clientid)
{
  DEBUGMSG(TRACE_DBG, ("rmuserdir\n"));
  char ent_name[2*NAME_MAX + 2] = {0};
  char perment_name[2*NAME_MAX + 2] = {0};

  if (! user_has_dir(clientid))
    {
      return 0;
    }

  snprintf(ent_name, 2*NAME_MAX + 1, DATA_DIR "/%d", clientid);
  snprintf(perment_name, 2*NAME_MAX + 1, PERM_DIR "/%d", clientid);
  rmr(ent_name);
  rmr(perment_name);
  return 0;
}

int rm_user_file(int clientid, char* filename)
{
  DEBUGMSG(TRACE_DBG, ("rm_user_file\n"));
  char ent_name[3*NAME_MAX + 3] = {0};
  char perment_name[3*NAME_MAX + 3] = {0};

  if (strstr(filename, "..") != NULL)
    {
      log_message("Filename with '..', refusing");
      return -1;
    }

  snprintf(ent_name, 3*NAME_MAX + 2, DATA_DIR "/%d/%s", clientid, filename);
  snprintf(perment_name, 3*NAME_MAX + 2, PERM_DIR "/%d/%s", clientid, filename);
  
  unlink(perment_name);
  return unlink(ent_name);
}

char* get_user_file(int clientid, char* filename, int* filesize)
{
  DEBUGMSG(TRACE_DBG, ("get_user_file\n"));
  char ent_name[3*NAME_MAX + 3] = {0};
  FILE* f;
  long fsize;
  char *outbuf;

  if (strstr(filename, "..") != NULL)
    {
      log_message("Filename with '..', refusing");
      return NULL;
    }

  snprintf(ent_name, 3*NAME_MAX + 2, DATA_DIR "/%d/%s", clientid, filename);
  
  if ((f = fopen(ent_name, "r")) == NULL)
    {
      log_message("No such file %s", snprintf);
      return NULL;
    }

  fseek(f, 0, SEEK_END);
  fsize = ftell(f);
  fseek(f, 0, SEEK_SET);
  outbuf = xmalloc(fsize + 1);
  bzero(outbuf, fsize + 1);
  fread(outbuf, 1, fsize, f);
  fclose(f);

  if (filesize != NULL)
    {
      *filesize = fsize;
    }

  return outbuf;
}

int put_user_file(int clientid, char* filename, char* content)
{
  DEBUGMSG(TRACE_DBG, ("put_user_file\n"));
  char ent_name[3*NAME_MAX + 3] = {0};
  FILE* f;

  mk_user_dir(clientid);

  if (strstr(filename, "..") != NULL)
    {
      log_message("Filename with '..'");
      return -1;
    }

  snprintf(ent_name, 3*NAME_MAX + 2, DATA_DIR "/%d/%s", clientid, filename);
  
  DEBUGMSG(TRACE_DBG, ("put_user_file: fopen\n"));
  if ((f = fopen(ent_name, "w")) == NULL)
    {
      log_message("Failed opening %s for write", snprintf);
      return -1;
    }

  if (content != NULL)
    fputs(content, f);
  fclose(f);
  return 0;
}

/*
flags:
UF_PERM_ADD <- dodaj do ju istniejcych praw
UF_PERM_SET <- ustaw na wybran now warto

readerid:
UF_READER_ANY <- dostp dla wszystkich
UF_READER_NONE <- dostp dla nikogo
*/

#define UF_PERM_ADD 1
#define UF_PERM_SET 3

#define UF_READER_NONE 0
#define UF_READER_ANY 1

int reader_has_perm(int clientid, char* filename, int readerid)
{
  DEBUGMSG(TRACE_DBG, ("reader_has_perm\n"));
  char ent_name[3*NAME_MAX + 3] = {0};
  char perment_name[3*NAME_MAX + 3] = {0};
  int allowed;
  FILE* f;

  if (strstr(filename, "..") != NULL)
    {
      log_message("Filename with '..'");
      return 0;
    }

  snprintf(ent_name, 2*NAME_MAX + 1, DATA_DIR "/%d/%s", clientid, filename);
  if ((f = fopen(ent_name, "r")) == NULL)
    {
      return 0;
    }
  fclose(f);

  snprintf(perment_name, 2*NAME_MAX + 1, PERM_DIR "/%d/%s", clientid, filename);

  if ((f = fopen(perment_name, "r")) == NULL)
    {
      return 1; // ma prawa, bo plik nieistnieje
    }
  else
    {
      while (fscanf(f, "%d", &allowed) == 1)
	{
	  if (readerid == allowed)
	    {
	      fclose(f);
	      return 1;
	    }
	}
      fclose(f);
    }

  return 0;
}

int user_file_set_perm(int clientid, char* filename, int readerid, int flag)
{
  DEBUGMSG(TRACE_DBG, ("user_file_set_perm\n"));
  char ent_name[3*NAME_MAX + 3] = {0};
  char perment_name[3*NAME_MAX + 3] = {0};
  char faccess_mode[2];
  FILE* f;

  if (strstr(filename, "..") != NULL)
    {
      log_message("Filename with '..'");
      return -1;
    }

  snprintf(ent_name, 2*NAME_MAX + 1, DATA_DIR "/%d/%s", clientid, filename);
  if ((f = fopen(ent_name, "r")) == NULL)
    {
      return -1;
    }
  fclose(f);

  snprintf(perment_name, 2*NAME_MAX + 1, PERM_DIR "/%d/%s", clientid, filename);

  if (readerid == UF_READER_ANY)
    {
      unlink(perment_name);
      return 0;
    }

  if (flag == UF_PERM_SET)
    {
      strcpy(faccess_mode, "w");
    }
  else
    {
      strcpy(faccess_mode, "a");
    }
  
  if ((f = fopen(perment_name, faccess_mode)) == NULL)
    {
      log_message("Can't set perms for file %s (fopen w)", perment_name);
      return -1;
    }

  if (readerid != UF_READER_NONE)
    {
      fprintf(f, "%d\n", readerid);
    }
  
  fclose(f);

  return 0;
}

int user_account_share()
{
  serv_mem* mem = shared_space_location(NULL);
  ASSERT(mem!=NULL,);
  user_info* ui = user_info_location(NULL);
  ASSERT(ui!=NULL,);

  if (ui->filescount < mem->config_vars[MAX_FILES_PERCLIENT])
    {
      p(SERV_MEM_SEM);
      mem->filescount++;
      ui->filescount++;
      v(SERV_MEM_SEM);
      return 1;
    }
  else
    return 0;
}

void user_account_unshare(int count)
{
  serv_mem* mem = shared_space_location(NULL);
  ASSERT(mem!=NULL,);
  user_info* ui = user_info_location(NULL);
  ASSERT(ui!=NULL,);

  p(SERV_MEM_SEM);
  mem->filescount = count;
  ui->filescount -= count;
  v(SERV_MEM_SEM);
}

int user_account_get(int filesize)
{
  serv_mem* mem = shared_space_location(NULL);
  ASSERT(mem!=NULL,);
  user_info* ui = user_info_location(NULL);
  ASSERT(ui!=NULL,);

  if (ui->transfer + filesize <= mem->config_vars[MAX_DOWNBYTES_PERCLIENT])
    {
      p(SERV_MEM_SEM);
      mem->transfer += filesize;
      ui->transfer += filesize;
      v(SERV_MEM_SEM);
      return 1;
    }
  else
    return 0;
}

/**************************************
 * shared space functions
 **************************************/

/* Tworzy obszar pamici wspdzielonej do przechowywania informacji
   o uytkownikach. W przypadku niepowodzenia wywouje safe_exit(SIGTERM)

   Zwraca wskanik do nowo utworzonego obszaru pamici. W strukturze uzupenia
   pole shmid, a pozostae zostawia wyzerowane
 */
users_info_mem* allocate_users_info_space(void)
{
  DEBUGMSG(TRACE_DBG, ("allocate_users_info_space\n"));
  DEBUGMSG(MEM_DBG, ("allocating users info space...\n"));
  int shmid = shmget(IPC_PRIVATE, sizeof(users_info_mem), 0644|IPC_CREAT);
  users_info_mem* mem = NULL;

  if (shmid == -1)
    {
      log_message("allocate_users_info_space shmget");
      safe_exit(SIGTERM);
    }

  mem = shmat(shmid, 0, 0);

  if ((int)mem == -1)
    {
      log_message("allocate_users_info_space shmat");
      safe_exit(SIGTERM);
    }
  bzero(mem, sizeof(users_info_mem));
  mem->shmid = shmid;
  
  return mem;
}

/* Zwalnia wspdzielony obszar pamici (razem z kolejnymi
   dodatkowo zaalokowanymi obszarami), jeli nie jest on NULL,
   Ewentualne bdy s wywietlane, ale nie powoduj zatrzymania
   programu.
*/
void unallocate_users_info_space(users_info_mem* mem)
{
  DEBUGMSG(TRACE_DBG, ("unallocate_users_info_space\n"));
  DEBUGMSG(MEM_DBG, ("unallocating users info space...\n"));
  int shmid;
  if (mem != NULL)
    {
      unallocate_users_info_space(mem->next_users);

      shmid = mem->shmid;
      if (shmdt(mem) == -1)
	{
	  log_message("unallocate_users_info_space shmdt");
	}
      if (shmctl(shmid, IPC_RMID, NULL) == -1)
	{
	  log_message("unallocate_users_info_space shmctl");
	}
    }
}

/* Wskanik do pamici wspdzielonej serwera.

   Jeli argument!=NULL, zapamituje wskanik do wspdzielonej pamici.
   Jeli argument==NULL, zwraca wskanik do tej pamici.
   Jeli argument==-1, ustawia wskanik na NULL
   Dziaa podobnie do zmiennej globalnej, wykorzystywane przy safe_exit,
   by atwo znale obszar pamici wspdzielonej do zwolnienia.
 */
serv_mem* shared_space_location(serv_mem* pointer)
{
  DEBUGMSG(TRACE_DBG, ("shared_space_location\n"));
  static serv_mem* mem_ptr = NULL;
  if (pointer != NULL)
    mem_ptr = pointer;
  if ((int)pointer == -1)
    mem_ptr = NULL;
  return mem_ptr;
}

/* Wskanik do obszaru z obsugiwanym uytkownikiem (dla procesw potomnych) 
 */
user_info* user_info_location(user_info* pointer)
{
  DEBUGMSG(TRACE_DBG, ("user_info_location\n"));
  static user_info* mem_ptr = NULL;
  if (pointer != NULL)
    mem_ptr = pointer;
  if ((int)pointer == -1)
    mem_ptr = NULL;
  return mem_ptr;
}

/* Tworzy pami wspdzielon, to znaczy obszar do przechowywania zmiennych
   konfiguracyjnych serwera (struktur serv_mem), oraz obszar do przechowywania
   informacji o kontach uytkownikw
*/
serv_mem* make_shared_space(void)
{
  DEBUGMSG(TRACE_DBG, ("make_shared_space\n"));
  DEBUGMSG(GENERAL_DBG, ("making shared space...\n"));
  int shmid = shmget(IPC_PRIVATE, sizeof(serv_mem), 0644|IPC_CREAT);
  serv_mem* mem = NULL;

  if (shmid == -1)
    {
      log_message("make_shared_space shmget");
      safe_exit(SIGTERM);
    }

  mem = shmat(shmid, 0, 0);
  DEBUGMSG(MEM_DBG, ("attaching shared memory at %x\n", (unsigned int)mem));
  
  if ((int)mem == -1)
    {
      log_message("make_shared_space shmat");
      safe_exit(SIGTERM);
    }
  bzero(mem, sizeof(serv_mem));
  mem->servpid = getpid();
  mem->shmid = shmid;
  mem->users = allocate_users_info_space();
  
  mem->semid = semget(IPC_PRIVATE, 2, 0644|IPC_CREAT|IPC_EXCL);
  if (mem->semid == -1)
    {
      log_message("make_shared_space semget");
      safe_exit(SIGTERM);
    }
  semctl(mem->semid, SERV_MEM_SEM, SETVAL, 1);
  mem->config_vars[MAX_CLIENTS] = MAX_CLIENTS_DEFAULT;
  mem->config_vars[MAX_FILES_PERCLIENT] = MAX_FILES_PERCLIENT_DEFAULT;
  mem->config_vars[MAX_DOWNBYTES_PERCLIENT] = MAX_DOWNBYTES_PERCLIENT_DEFAULT;
  mem->msgid = -1;

  mkdir(SERVER_DIR, 0755);
  mkdir(DATA_DIR, 0755);
  mkdir(PERM_DIR, 0755);

  return shared_space_location(mem);
}

/* Usuwa obszar pamici wspdzielonej, razem z obszarami pamici na 
   dane o uytkownikach i ustawia wskanik do tego obszaru na NULL.
   Ewentualne bdy s wywietlane, ale nie powoduj zatrzymania
   programu.
 */
void remove_shared_space(void)
{
  DEBUGMSG(TRACE_DBG, ("remove_shared_space\n"));
  DEBUGMSG(GENERAL_DBG, ("removing shared space...\n"));
  int shmid;
  serv_mem* mem = shared_space_location(NULL);
  
  if (mem != NULL)
    {
      unallocate_users_info_space(mem->users);
      
      if (semctl(mem->semid, 0, IPC_RMID))
	{
	  log_message("remove_shared_space semctl");
	}

      shmid = mem->shmid;

      if (shmdt(mem) == -1)
	{
	  log_message("remove_shared_space shmdt");
	}
      if (shmctl(shmid, IPC_RMID, NULL) == -1)
	{
	  log_message("remove_shared_space shmctl");
	}

      shared_space_location((serv_mem*)-1);
    }

  rmr(SERVER_DIR);
}

/**************************************
 * semaphore operations
 **************************************/

/* prosta operacja do zmiany wartoci na jednym semaforze,
  z debugowaniem. zalecane korzystanie z V i P, jeli semafor
  bdzie zmieniany o 1. Obsuguje przypadek, gdy operacja na
  semaforze zwrci bd.
*/
void semchange(int semnum, int changeval)
{
  DEBUGMSG(TRACE_DBG, ("semchange\n"));
  serv_mem* mem = shared_space_location(NULL);
  struct sembuf sops;
  DEBUGMSG(SEM_DBG, ("semaphore %d change by %d\n", semnum, changeval));
  ASSERT(mem!=NULL,);

  sops.sem_num = semnum;
  sops.sem_op = changeval;
  if (semop(mem->semid, &sops, 1) == -1)
    {
      log_message("semchange semop");
      safe_exit(SIGTERM);
    }
}

/* podnie semafor 
*/
void v(int semnum)
{
  DEBUGMSG(TRACE_DBG, ("v\n"));
  semchange(semnum, +1);
}

/* opu semafor
 */
void p(int semnum)
{
  DEBUGMSG(TRACE_DBG, ("p\n"));
  semchange(semnum, -1);
}

/* Zmienia opcje konfiguracyjne serwera. Operacja wymaga wycznoci, 
   do tego jest wykorzystywany semafor SERV_MEM_SEM. Jeli zmiana 

   Rejestruje nowe poczenie i zwraca wskanik do niego w pamici serwera. 
   Obnia poziom semaforu USERSCOUNT_SEM.
   W przypadku bdu zwraca NULL
 */
int register_connection(user_info** newui)
{
  DEBUGMSG(SESSION_DBG|TRACE_DBG, ("register_connection\n"));
  serv_mem* mem = shared_space_location(NULL);
  users_info_mem* users;
  complete_msg msg;  
  int cnum = 0;
  ASSERT(mem!=NULL,);
  user_info* newtmp = NULL;

  /* oczekiwanie na wiadomo od niezalogowanego z chci zalogowania */
  receive_message(mem->msgid, CLIENTID_LOGIN_REQUEST, CMD_LOGIN, &msg, 0);

  log_message("clientid:%d LOGIN user %s with id %s", msg.clientid, msg.argument1, msg.argument2);

  *newui = NULL;

  /* weryfikacja argumentw */
  if (atoi(msg.argument2) == 1 || atoi(msg.argument2) == 0) // clientid!=0 i clientid!=1
    {
      log_message("Incorrect clientid.");
      return 1;
    }
  msg.clientid = atoi(msg.argument2);

  if (msg.argument1==NULL)
    {
      answer_error(&msg, "Username required.");
      return 1;
    }

  users = mem->users;
  ASSERT(users!=NULL,);

  p(SERV_MEM_SEM); // sekcja krytyczna w zwizku z operacjami na SERV_MEM

  if (mem->clients >= mem->config_vars[MAX_CLIENTS])
    {
      answer_error(&msg, "Users limit. Please try again later.");
      v(SERV_MEM_SEM);
      free_message(&msg);
      return 1;
    }

  while (cnum < USERS_INFO_SIZE)
    {
      if (strcmp(users->users[cnum].name, msg.argument1) == 0)
	{
	  answer_error(&msg, "This user is already logged in."); // 2 wyjscie z sekcji kryt.
	  v(SERV_MEM_SEM);
	  free_message(&msg);
	  return 1;
	}

      if ((users->users[cnum].clientid == 0) && (newtmp == NULL))
	{
	  newtmp = &(users->users[cnum]);
	}

      cnum++;
      if ((cnum == USERS_INFO_SIZE) && (users->next_users == NULL) && (newtmp == NULL))
	{
	  users->next_users = allocate_users_info_space();
	  users = users->next_users;
	  cnum = 0;
	}
    }

  mem->clients++;
  bzero(newtmp, sizeof(user_info));
  newtmp->clientid = atoi(msg.argument2);
  strncpy(newtmp->name, msg.argument1, 16);
  if (strcmp(msg.argument1, "admin") == 0)
    {
      newtmp->admin = 1;
    }
  v(SERV_MEM_SEM); // koniec sekcji krytycznej (w srodku jest 2 wyjscie)
  *newui = newtmp;
  answer_success(&msg);
  free_message(&msg);

  return 1;
}

/* Zwalnia miejsce np. po wylogowaniu, lub wykickowaniu, lub przed
   zamkniciem serwera. Zwalnia miejsce cnum w tablicy mem->users i
   podnosi semafor USERSCOUNT_SEM
 */
void unregister_connection(user_info* ui)
{
  serv_mem* mem = shared_space_location(NULL);
  ASSERT(mem!=NULL,);
  DEBUGMSG(TRACE_DBG, ("unregister_connection\n"));
  ASSERT(ui!=NULL,);
  DEBUGMSG(SESSION_DBG, ("unregister_connection\n"));

  log_message("Unregistering clientid:%d", ui->clientid);

  user_account_unshare(ui->filescount);
  rm_user_dir(ui->clientid);  

  p(SERV_MEM_SEM);
  mem->clients--;
  v(SERV_MEM_SEM);

  bzero(ui, sizeof(user_info));
}

/* Zmienia opcje konfiguracyjne serwera. Operacja wymaga wycznoci, 
   do tego jest wykorzystywany semafor SERV_MEM_SEM. Jeli zmiana 
   dotyczy maks.liczby uytkownikw, zmieniana jest te maksymalna 
   warto semafora USERSCOUNT_SEM
 */
void set_config_var(int optnum, int newval)
{
  DEBUGMSG(TRACE_DBG, ("set_config_var\n"));
  serv_mem* mem = shared_space_location(NULL);
  ASSERT(mem!=NULL,);
  DEBUGMSG(GENERAL_DBG|SESSION_DBG,("Remote setting var %d to %d\n", optnum, newval));

  log_message("Setting config_var[%d] = %d", optnum, newval);
  p(SERV_MEM_SEM);
  mem->config_vars[optnum] = newval;
  v(SERV_MEM_SEM);
}

int username2clientid(char *username)
{
  serv_mem* mem = shared_space_location(NULL);
  ASSERT(mem!=NULL,);
  users_info_mem* users = mem->users;
  ASSERT(users!=NULL,);
  int i=0, clientid=0;

  while (clientid==0 && i<USERS_INFO_SIZE)
    {
      if ((users->users[i].clientid!=0) && (strcmp(users->users[i].name, username) == 0))
	{
	  clientid = users->users[i].clientid;
	}

      i++;
      if ((i==USERS_INFO_SIZE)  && (users->next_users!=NULL))
	{
	  i=0;
	  users = users->next_users;
	}
    }
 
  return clientid;
}

/* Alokuje miejsce na zwracane username - trzeba je potem zwolni!
 */
char* clientid2username(int clientid)
{
  serv_mem* mem = shared_space_location(NULL);
  ASSERT(mem!=NULL,);
  users_info_mem* users = mem->users;
  ASSERT(users!=NULL,);
  int i=0;
  char* username = NULL;

  while (username==NULL && i<USERS_INFO_SIZE)
    {
      if (users->users[i].clientid == clientid)
	{
	  username = (char*) xmalloc(strlen(users->users[i].name) + 1);
	  bzero(username, strlen(users->users[i].name) + 1);
	  strcpy(username, users->users[i].name);
	}

      i++;
      if ((i==USERS_INFO_SIZE)  && (users->next_users!=NULL))
	{
	  i=0;
	  users = users->next_users;
	}
    }
  
  return username;
}

/**************************************
 * Polecenia
 **************************************/

void list_shares(complete_msg* msg)
{
  serv_mem* mem = shared_space_location(NULL);
  ASSERT(mem!=NULL,);
  users_info_mem* users = mem->users;
  ASSERT(users!=NULL,);
  char *buf_ptr;
  int bufsize;
  int i = 0;

  log_message("clientid:%d LIST", msg->clientid);
  bufsize = mem->config_vars[MAX_CLIENTS] * (2*PROTO_ARGSIZE +2);
  buf_ptr = (char*)xmalloc(bufsize + 1);
  bzero(buf_ptr, bufsize+1);
  msg->messagebody = buf_ptr;

  while (i<USERS_INFO_SIZE)
    {
      if (users->users[i].clientid!=0)
	{
	  if (msg->argument1 == NULL || strncmp(msg->argument1, users->users[i].name, PROTO_ARGSIZE) == 0)
	  ls(users->users[i].clientid, users->users[i].name, &buf_ptr, &bufsize);
	}

      i++;
      if ((i==USERS_INFO_SIZE)  && (users->next_users!=NULL))
	{
	  i=0;
	  users = users->next_users;
	}
    }
  
  send_message(msg->clientid, msg);
  xfree(msg->messagebody);
  msg->messagebody = NULL;
}

void cmd_get(complete_msg* msg)
{
  DEBUGMSG(TRACE_DBG, ("cmd_get\n"));
  int fileowner_id = username2clientid(msg->argument1);
  char* buf;
  int filesize;

  log_message("clientid:%d GET file %s from %s", msg->clientid, msg->argument2, msg->argument1);

  if (fileowner_id == 0)
    {
      answer_error(msg, "User not found");
      return;
    }

  buf = get_user_file(fileowner_id, msg->argument2, &filesize);
  if (buf == NULL)
    {
      answer_error(msg, "File not found");
      return;
    }

  if (! reader_has_perm(fileowner_id, msg->argument2, msg->clientid))
    {
      answer_error(msg, "Access denied");
      return;
    }

  if (! user_account_get(filesize))
    {
      answer_error(msg, "File size exceeds your transfer limit");
      return;
    }

  msg->messagebody = buf;
  send_message(msg->clientid, msg);
  msg->messagebody = NULL;
  xfree(buf);
}

void cmd_share(complete_msg* msg)
{
  DEBUGMSG(TRACE_DBG, ("cmd_share\n"));
  int accessor_id;

  if (msg->argument2 != NULL)
    log_message("clientid:%d SHARE file %s for %s", msg->clientid, msg->argument1, msg->argument2);
  else
    log_message("clientid:%d SHARE file %s for ALL", msg->clientid, msg->argument1);

  if (! user_account_share())
    {
      answer_error(msg, "Files count limit exceeded");
      return;
    }

  if (put_user_file(msg->clientid, msg->argument1, msg->messagebody) == -1)
    {
      answer_error(msg, "Share failed");
      return;
    }
  
  if (msg->argument2 != NULL)
    {
      accessor_id = username2clientid(msg->argument2);
      if (accessor_id == 0)
	{
	  answer_error(msg,"User unknown, allowing only for You");
	  user_file_set_perm(msg->clientid, msg->argument1, msg->clientid, UF_PERM_ADD);
	}
      else
	{
	  user_file_set_perm(msg->clientid, msg->argument1, accessor_id, UF_PERM_ADD);
	}
    }
  else
    {
      user_file_set_perm(msg->clientid, msg->argument1, UF_READER_ANY, UF_PERM_SET);
    }
  answer_success(msg);
}

void list_users(complete_msg* msg)
{
  serv_mem* mem = shared_space_location(NULL);
  ASSERT(mem!=NULL,);
  users_info_mem* users = mem->users;
  ASSERT(users!=NULL,);
  char* buf_ptr;
  int i=0;

  if (msg->argument1 != NULL)
    log_message("clientid:%d WHO %s", msg->clientid, msg->argument1);
  else    
    log_message("clientid:%d WHO", msg->clientid);

  msg->messagebody = buf_ptr = (char*)xmalloc(mem->config_vars[MAX_CLIENTS] * (PROTO_ARGSIZE+1)); 

  bzero(buf_ptr, mem->config_vars[MAX_CLIENTS] * (PROTO_ARGSIZE+1));
  while (i<USERS_INFO_SIZE)
    {
      if (users->users[i].clientid!=0)
	{
	  buf_ptr = (char*)((unsigned int)buf_ptr + (unsigned int)sprintf(buf_ptr, "%s\n", users->users[i].name));
	}

      i++;
      if ((i==USERS_INFO_SIZE)  && (users->next_users!=NULL))
	{
	  i=0;
	  users = users->next_users;
	}
    }

  send_message(msg->clientid, msg);
}

void show_stat(complete_msg* msg)
{
  serv_mem *mem = shared_space_location(NULL);
  user_info *ui = user_info_location(NULL);
  ASSERT(ui!=NULL,);
  ASSERT(mem!=NULL,);
  users_info_mem* users = mem->users;
  char* buf;
  char* buf_ptr;
  int i=0;

  log_message("clientid:%d STAT", msg->clientid);  

  if (ui->admin)
    {
      buf = buf_ptr = (char*) xmalloc(3*20 + mem->clients*(PROTO_ARGSIZE + 20));
      bzero(buf_ptr, 3*20 + mem->clients*(PROTO_ARGSIZE + 20));
      buf_ptr =  (char*)((unsigned int)buf_ptr + (unsigned int)sprintf(buf_ptr, "x %d %d\ny %d %d\nz %d %d\n", 
		  mem->clients,      mem->config_vars[MAX_CLIENTS],
		  mem->filescount,   mem->config_vars[MAX_FILES_PERCLIENT],
		  mem->transfer,     mem->config_vars[MAX_DOWNBYTES_PERCLIENT]));
 
      while (i<USERS_INFO_SIZE)
	{
	  if (users->users[i].clientid!=0)
	    {
	      buf_ptr = (char*)((unsigned int)buf_ptr + (unsigned int)sprintf(buf_ptr, "%s %d %d\n", 
									      users->users[i].name,
									      users->users[i].filescount,
									      users->users[i].transfer));
	    }
	  
	  i++;
	  if ((i==USERS_INFO_SIZE)  && (users->next_users!=NULL))
	    {
	      i=0;
	      users = users->next_users;
	    }
	}
    }
  else
    {
      buf = (char*) xmalloc(30);
      bzero(buf, 30);
      snprintf(buf, 30, "%d %d\n%d %d", 
	       ui->filescount, mem->config_vars[MAX_FILES_PERCLIENT],
	       ui->transfer, mem->config_vars[MAX_DOWNBYTES_PERCLIENT]);
    }

  msg->messagebody = buf;
  send_message(msg->clientid, msg);
  msg->messagebody = NULL;
}

void set(complete_msg* msg)
{
  user_info* ui = user_info_location(NULL);
  ASSERT(ui!=NULL,);
  int new_val;

  log_message("clientid:%d SET %s to %s", msg->clientid, msg->argument1, msg->argument2);

  if (ui->admin)
    {
      new_val = strtol(msg->argument2, NULL, 10);
      if (errno == EINVAL || errno == ERANGE)
	{
	  answer_error(msg, "Wrong value");
	}
      else
	{
	  log_message("SET %c to %d", msg->argument1[0], new_val);
	  switch (msg->argument1[0])
	    {
	    case ('x'):
	      set_config_var(MAX_CLIENTS, new_val);
	      answer_success(msg);
	      break;
	    case ('y'):
	      set_config_var(MAX_FILES_PERCLIENT, new_val);
	      answer_success(msg);
	      break;
	    case ('z'):
	      set_config_var(MAX_DOWNBYTES_PERCLIENT, new_val);
	      answer_success(msg);
	      break;
	    default:
	      answer_error(msg, "Unknown variable");
	    }
	}
    }
  else
    {
      answer_error(msg, "Operation not permitted.");
    }
}

void kick(complete_msg* msg)
{
  serv_mem *mem = shared_space_location(NULL);
  ASSERT(mem!=NULL,);
  users_info_mem *users;
  user_info* ui = user_info_location(NULL);
  ASSERT(ui!=NULL,);
  int i=0;

  log_message("clientid:%d KICK %s", msg->clientid, msg->argument1);

  users = mem->users;

  if (ui->admin)
    {
      while ((i<USERS_INFO_SIZE)  && (strcmp(msg->argument1, users->users[i].name)!=0))
	{
	  i++;
	  if ((i==USERS_INFO_SIZE)  && (users->next_users!=NULL))
	    {
	      i=0;
	      users = users->next_users;
	    }
	}
 
      if (i<USERS_INFO_SIZE)
	{
	  answer_success(msg);
	  if (users->users[i].clientid != msg->clientid)
	    { /* jesli zabijamy nie siebie */
	      send_message(users->users[i].clientid, msg);
	      kill(users->users[i].childpid, SIGTERM);
	    }
	  else
	    {
	      free_message(msg);
	      unregister_connection(user_info_location(NULL));
	      exit(0);
	    }
	}
      else
	{
	  answer_error(msg, "User not found");
	}
    }
  else
    {
      answer_error(msg, "Operation not permitted");
    }
}

/**************************************
 * EXTENDED PROTOCOL OPTIONS
 **************************************/

void allow(complete_msg* msg)
{
  int accessor_id;
  if (msg->argument2 == NULL)
    {
      if (user_file_set_perm(msg->clientid, msg->argument1, UF_READER_NONE, UF_PERM_SET) == -1)
	{
	  answer_error(msg, "Allow failed. Maybe there is no file, or other error occured");
	}
      else
	{
	  answer_success(msg);
	}
    }
  else
    {
      accessor_id = username2clientid(msg->argument2);
      
      if (accessor_id == 0)
	{
	  answer_error(msg, "User not found");
	}
      else
	{
	  if (user_file_set_perm(msg->clientid, msg->argument1, accessor_id, UF_PERM_ADD) == -1)
	    {
	      answer_error(msg, "Allow failed. Maybe there is no file, or other error occured");
	    }
	  else
	    {
	      answer_success(msg);
	    }
	}
    }
}

void kickall(complete_msg* msg)
{
  serv_mem *mem = shared_space_location(NULL);
  ASSERT(mem!=NULL,);
  users_info_mem *users;
  user_info* ui = user_info_location(NULL);
  ASSERT(ui!=NULL,);
  int i=0;

  log_message("clientid:%d KICKALL", msg->clientid);

  users = mem->users;
  
  if (ui->admin)
    {
      while (i<USERS_INFO_SIZE)
	{
	  if ((users->users[i].clientid != 0) && (users->users[i].clientid != msg->clientid))
	    { /* siebie nie zabijamy */
	      msg->command = CMD_KICK;    
	      msg->argument1 = clientid2username(users->users[i].clientid);
	      send_message(users->users[i].clientid, msg);
	      xfree(msg->argument1);
	      msg->argument1 = NULL;
	      kill(users->users[i].childpid, SIGTERM);
	    }

	  i++;
	  if ((i==USERS_INFO_SIZE)  && (users->next_users!=NULL))
	    {
	      i=0;
	      users = users->next_users;
	    }
	}

      answer_success(msg);
    }
  else
    {
      answer_error(msg, "Operation not permitted");
    }
}

void shutdown(complete_msg* msg)
{
  user_info* ui = user_info_location(NULL);
  ASSERT(ui!=NULL,);
  serv_mem *mem = shared_space_location(NULL);
  ASSERT(mem!=NULL,);
  
  log_message("clientid:%d SHUTDOWN", msg->clientid);
  
  if (ui->admin)
    {
      answer_success(msg);
      kill(mem->servpid, SIGTERM);
    }
  else
    {
      answer_error(msg, "Operation not permitted.");
    }  
}

void unshare(complete_msg* msg)
{
  if (rm_user_file(msg->clientid, msg->argument1) != -1)
    {
      user_account_unshare(1);
      answer_success(msg);
    }
  else
    {
      answer_error(msg, "Unshare failed");
    }
}

/**************************************
 * session functions
 **************************************/

void handle_session(user_info* ui)
{
  DEBUGMSG(TRACE_DBG, ("handle_session\n"));
  serv_mem* mem = shared_space_location(NULL);
  complete_msg msg;

  user_info_location(ui);

  ASSERT(ui!=NULL,);
  DEBUGMSG(SESSION_DBG,("new session with user clientid%d\n", ui->clientid));
  ASSERT(mem!=NULL,);

  log_message("New session for clientid:%d", ui->clientid);

  p(SERV_MEM_SEM); // jedyna operacja, ktora musi byc wykonywana w trybie wylacznosci,
                   // pozniej juz, proces ma na wylacznosc ten kawalek pamieci (do zapisu)
                   // inni conajwyzej beda odczytywac
  ui->childpid = getpid();
  v(SERV_MEM_SEM);

  do {
    free_message(&msg);
    receive_message(mem->msgid, ui->clientid, 0, &msg, KEEP_UNWANTED);

    switch (msg.command)
      {
      case (CMD_LOGIN):
	answer_error(&msg, "Already logged in.");
	break;
      case (CMD_LOGOUT):
	log_message("clientid:%d LOGOUT", msg.clientid);
	answer_success(&msg);
	break;
      case (CMD_SHARE):
	cmd_share(&msg);
	break;
      case (CMD_GET):
	cmd_get(&msg);
	break;
      case (CMD_LIST):
	list_shares(&msg);
	break;
      case (CMD_WHO):
	list_users(&msg);
	break;
      case (CMD_STAT):
	show_stat(&msg);
	break;
      case (CMD_KICK):
	kick(&msg);
	break;
      case (CMD_SET):
	set(&msg);
	break;	
      case (ECMD_ALLOW):
	allow(&msg);
	break;
      case (ECMD_SHUTDOWN):
	shutdown(&msg);
	break;
      case (ECMD_KICKALL):
	kickall(&msg);
	break;
      case (ECMD_UNSHARE):
	unshare(&msg);
	break;
      default:
	log_message("clientid:%d UNKNOWN COMMAND command:%d", msg.clientid, msg.command);
	answer_error(&msg, "Unknown command.");
      }

  } while (msg.command != CMD_LOGOUT);

  free_message(&msg);
  unregister_connection(user_info_location(NULL));
}

void close_sessions(void)
{
  DEBUGMSG(TRACE_DBG, ("close_sessions\n"));
  int cnum = 0;
  serv_mem* mem = shared_space_location(NULL);
  ASSERT(mem!=NULL,);
  DEBUGMSG(SESSION_DBG, ("closing sessions\n"));
  users_info_mem* users;

  users = mem->users;
  ASSERT(users!=NULL,);  

  while (cnum != USERS_INFO_SIZE)
    {  
      if (users->users[cnum].clientid != 0)
	{
	  DEBUGMSG(SESSION_DBG, ("killing process %d\n", users->users[cnum].childpid));
	  kill(users->users[cnum].childpid, SIGTERM);
	}

      cnum++;
      if ((cnum == USERS_INFO_SIZE) && (users->next_users != NULL))
	{
	  users = users->next_users;
	  cnum = 0;
	}
    }
}

/**************************************
 * main
 **************************************/
int main(int argc, char* argv[])
{
  serv_mem* mem = NULL;
  user_info* new_user;
  pid_t procpid;
  int fd;
  char str[10] = {0};

  procpid = fork();
  
  if (procpid<0)
    {
      log_message("Fork failed");
      exit(1);
    }
  if (procpid>0)
    {
      exit(0);
    }

  
  printf("66301_server: serwer sieci IPC1500. Autor: Jacek Pospychala (inf66301)\n"
	 "wyjscie: kill `cat " LOCK_FILE "`\n"
	 "log: tail -f " LOG_FILE "\n");
  setsid();

  for (fd=getdtablesize();fd>=0;fd--) 
    {
      close(fd);
    }

  fd=open("/dev/null",O_RDWR); /* open stdin */
  dup(fd); /* stdout */
  dup(fd); /* stderr */

  fd=open(LOCK_FILE,O_RDWR|O_CREAT,0640);
  if (fd<0) 
    {
      log_message("Can't open lock file (\"" LOCK_FILE "\")");
      exit(1);
    }
  if (lockf(fd,F_TLOCK,0)<0) 
    { 
      log_message("Server is already running");
      exit(0); 
    }
  
  sprintf(str,"%d\n",getpid());
  write(fd,str,strlen(str)); /* record pid to lockfile */
  
  umask(027);
  chdir(RUNNING_DIR);

  log_message("Ready... Steady... Go!");

  mem = make_shared_space();
  mem->msgid = create_queue(SERVERID);

  signal(SIGTERM, &safe_exit);
  signal(SIGINT, &safe_exit);
  signal(SIGHUP, &safe_exit);

  while (register_connection(&new_user))
    {
      if ((new_user != NULL) && (fork()==0))
	{
	  handle_session(new_user);
	  exit(0);
	  }
    }

  safe_exit(SIGTERM);
  return 0;
}
