/* concurrencytest
   Jacek Pospychala <tri10o@siat.pl>
*/

#include <ncurses.h>
#include <pthread.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ipc.h>
#include <sys/sem.h>

// wyswietlanie: D-dumb, B-before, A-after, P-print
#define XD(lnum,a) a
#define XB(lnum,a) mvprintw(offt+lnum, (w/n)*i +1, "%3d. %s", lnum, #a); refresh(); state(); hang(offt+lnum, i); state(); a
#define XA(lnum,a) a; mvprintw(offt+lnum, (w/n)*i +1, "%3d. %s", lnum, #a); refresh(); hang(offt+lnum, i); state();
#define XP(lnum,a) mvprintw(offt+lnum, (w/n)*i +1, "%3d. %s", lnum, #a); refresh(); state(); hang(offt+lnum, i); state();
int w, h, offt=1;
int n = 0;
pthread_mutex_t* m;
void hang(int line, int i)
{
  mvprintw(line, ((w/n)*i), ">"); refresh();
  pthread_mutex_lock(&m[i]);
  mvprintw(line, ((w/n)*i), " "); refresh();
}

// operacje samaforowe
int semid;
void P(int semnum)
{
  struct sembuf s = {0};
  s.sem_num = semnum;
  s.sem_op = -1;
  semop(semid, &s, 1);
}
void V(int semnum)
{
  struct sembuf s = {0};
  s.sem_num = semnum;
  s.sem_op = 1;
  semop(semid, &s, 1);
}

void (*algorytm)(int) = NULL;
void (*algorytmy[5])(int) = {NULL};
void (*state)() = NULL;

// TWOJ ALGORYTM - HOW-TO:
// na poczatku deklaracje zmiennych, z poczatkowymi wartosciami
// nastepnie funkcja z algorytmem dla i-tego procesu. ta funkcja bedzie
// uruchomiona n razy jednoczesnie, a uzytkownik bedzie mogl w kazdej funkcji
// wykonywac instrukcje krok po kroku.
//
// funkcja musi miec argument typu int, o nazwie i. w nim bedzie numer procesu
// kazda linia funkcji musi byc zamknieta w jednym z makr: D, B, A, lub P
// wszystkie one przyjmuja te same argumenty:
//
// numer_linii, linia kodu algorytmu
//
// numer linii to bedzie zazwyczaj numer porzadkowy, zaczynajac od 0
// linia kodu programu to linia, ktora ma byc wykonana w danym kroku
//
// ktore makro uzyc? D, B, A czy P?
//
// D - powoduje tylko wykonanie danej linii, a nie wypisuje jej. stosowane przy
//     liniach z deklaracja funkcji, gdzie jeszcze nie mozna nic wypisac, bo jestesmy
//     poza funkcja i nie mozemy wykonywac zadnych polecen
// B - Before - najpierw wypisywana jest tresc linii kodu, a wykonywany jest on PO potwierdzeniu klawiszem
//     wykorzystywane najczesciej w typowym kodzie: "Najpierw pokaz co chcesz zrobic, a jak uzytkownik sie
//     zgodzi, to to zrob."
// A - After - najpierw wykonywana jest linia kodu, a potem jest to wypisywane na ekranie. Uzywane glownie
//     przy nawiasach otwierajacych, gdy nie chcemy zamieszac porzadku wykonywania algorytmu przez funkcje
//     wypisujace kroki na ekranie
// P - samo wypisanie linii kodu na ekranie. Uzywane do linii, ktore nie robia nic istotnego z punktu widzenia
//     algorytmu, czyli wywolania w stylu SekcjaKrytyczna(), RobCos(), itp.
//
// na koncu funkcja wypisujaca stan wszystkich interesujacych zmiennych
// w dalszej czesci programu nalezy dodac fragment dotyczacy Twojego algorytmu
// w razie klopotow, zawsze warto zobaczyc, jaki kod jest generowany przez preprocesor (cpp).

// ALGORYTM 1
int processNumber = 1;
XD(0, void alg1_processi(int i)                   ) 
XA(1, {                                           ) 
XB(2,   while (TRUE)                              ) 
XA(3,     {                                       ) 
XB(4,       while (processNumber==((i+1) %2))     ) 
XA(5,         {                                   ) 
XB(6,         }                                   )
XP(7,       criticalSection();                    ) 
XB(8,       processNumber=(i+1)%2;                ) 
XP(9,       otherStuff();                         ) 
XB(10,     }                                      ) 
XB(11,}                                           ) 
void arg1_state() { mvprintw(0, 0, "processNumber=%d", processNumber); refresh(); }

// ALGORYTM 2
int PIinside[2] = {FALSE, FALSE};
XD(0,  void alg2_processi(int i)                  )
XA(1, {                                           ) 
XB(2,   while (TRUE)                              ) 
XA(3,     {                                       ) 
XB(4,       while (PIinside[(i+1)%2])             ) 
XA(5,         {                                   ) 
XB(6,         }                                   )
XB(7,       PIinside[i]=TRUE;                     )
XP(8,       criticalSection();                    ) 
XB(9,       PIinside[i]=FALSE;                    ) 
XP(10,      otherStuff();                         ) 
XB(11,     }                                      ) 
XB(12,}                                           ) 
void arg2_state() { mvprintw(0, 0, "P0inside=%d P1inside=%d", PIinside[0], PIinside[1]); refresh(); }

// ALGORYTM 3
int PIWantsToEnter[2] = {FALSE, FALSE};
XD(0,  void alg3_processi(int i)                  )
XA(1, {                                           ) 
XB(2,   while (TRUE)                              ) 
XA(3,     {                                       )
XB(4,       PIWantsToEnter[i]=TRUE;               ) 
XB(5,       while (PIWantsToEnter[(i+1)%2])       ) 
XA(6,         {                                   ) 
XB(7,         }                                   )
XP(8,       criticalSection();                    ) 
XB(9,       PIWantsToEnter[i]=FALSE;              ) 
XP(10,      otherStuff();                         ) 
XB(11,     }                                      ) 
XB(12,}                                           ) 
void arg3_state() { mvprintw(0, 0, "P0WantsToEnter=%d P1WantsToEnter=%d", PIWantsToEnter[0], PIWantsToEnter[1]); refresh(); }

// ALGORYTM 4
// uzywa zmiennej z algorytmu 3 !
XD(0,  void alg4_processi(int i)                  )
XA(1, {                                           ) 
XB(2,   while (TRUE)                              ) 
XA(3,     {                                       )
XB(4,       PIWantsToEnter[i]=TRUE;               ) 
XB(5,       while (PIWantsToEnter[(i+1)%2])       ) 
XA(6,         {                                   ) 
XB(7,           PIWantsToEnter[i]=FALSE;          ) 
XP(8,           delay(random, freecycles);        ) 
XB(9,           PIWantsToEnter[i]=TRUE;           ) 
XB(10,         }                                  )
XP(11,       criticalSection();                   ) 
XB(12,       PIWantsToEnter[i]=FALSE;             ) 
XP(13,      otherStuff();                         ) 
XB(14,     }                                      ) 
XB(15,}                                           ) 
void arg4_state() { mvprintw(0, 0, "P0WantsToEnter=%d P1WantsToEnter=%d", PIWantsToEnter[0], PIWantsToEnter[1]); refresh(); }

// ALGORYTM DEKKERA
int favouredProcess;
// uzywa zmiennych z algorytmu 3 !
XD(0,  void dekker_processi(int i)                )
XA(1, {                                           ) 
XB(2,   while (TRUE)                              ) 
XA(3,     {                                       )
XB(4,       PIWantsToEnter[i]=TRUE;               ) 
XB(5,       while (PIWantsToEnter[(i+1)%2])       ) 
XA(6,         {                                   ) 
XB(7,           PIWantsToEnter[i]=FALSE;          ) 
XB(8,           while (favouredProcess==(i+1)%2)  ) 
XA(9,             {                               ) 
XB(10,            }                               ) 
XB(11,          PIWantsToEnter[i]=TRUE;           ) 
XB(12,         }                                  )
XP(13,       criticalSection();                   ) 
XB(14,       favouredProcess=(i+1)%2;             ) 
XB(15,       PIWantsToEnter[i]=FALSE;             ) 
XP(16,      otherStuff();                         ) 
XB(17,     }                                      ) 
XB(18,}                                           )
void dekker_state() { offt=2; mvprintw(0, 0, "favouredProcess=%d\n"
				       "P0WantsToEnter=%d P1WantsToEnter=%d", 
				       favouredProcess, PIWantsToEnter[0], PIWantsToEnter[1]); refresh(); }

// ALGORYTM DIJKSTRY
// shared
int *flag;
int turn = 0;
// local
int *test, *k, *other;
XD(0,void dijkstra_processi(int i)                )
XA(1,{                                            ) 
XB(2,   while (TRUE)                              ) 
XA(3,     {                                       )
XB(4,       L: flag[i] = 1;                       ) 
XB(5,       other[i] = turn;                      ) 
XB(6,       while (other[i]!=i)                   ) 
XA(7,         {                                   ) 
XB(8,           test[i] = flag[other[i]];         ) 
XB(9,           if (test[i]==0)                   ) 
XA(10,            {                               ) 
XB(11,              turn = i;                     ) 
XB(12,            }                               ) 
XB(13,          other[i] = turn;                  ) 
XB(14,         }                                  )
XB(15,       flag[i] = 2;                         )
XB(16,       for (k[i]=0; k[i]<n; k[i]++)         )
XA(17,            {                               )
XB(18,              if (k[i]!=i)                  ) 
XA(19,                 {                          ) 
XB(20,                  test[i] = flag[k[i]];     ) 
XB(21,                   if (test[i] == 2)        ) 
XA(22,                     {                      ) 
XB(23,                       goto L;              ) 
XB(24,                    }                       ) 
XB(25,                }                           ) 
XB(26,            }                               )
XP(27,       criticalSection();                   ) 
XB(28,       flag[i]=0;                           ) 
XP(29,       reminderSection();                   ) 
XB(30,     }                                      ) 
XB(31,}                                           )
void dijkstra_state() { 
  int i;
  offt=5;
  mvprintw(0, 0, "turn=%d", turn); 
  for (i=0; i<n; i++)
    {
      mvprintw(1, (w/n)*i, "%3d: flag=%d", i, flag[i]);
      mvprintw(2, (w/n)*i, "     test=%d", test[i]);
      mvprintw(3, (w/n)*i, "     other=%d", other[i]);
      mvprintw(4, (w/n)*i, "     k=%d", k[i]);
    }
  refresh(); 
}

// EXAM: ZADANIE Z BUSAMI
int x1=0, x2=1, x3=2, x4=3, x5=4;
XD(0, void busA(int i)                            ) 
XA(1, {                                           ) 
XB(2,   while (TRUE)                              ) 
XA(3,     {                                       ) 
XP(4,       wjazd na stare miasto                 ) 
XP(5,       przejazd przez 1                      ) 
XP(6,       przejazd przez 2                      )
XB(7,       P(x3);                                ) 
XP(8,       przejazd przez 3                      ) 
XB(9,       P(x4);                                ) 
XP(10,      przejazd przez 4                      ) 
XB(11,      V(x3);                                ) 
XP(12,      wyjazd ze starego miasta              ) 
XB(13,    }                                       ) 
XB(14,}                                           ) 

XD(0, void busB(int i)                            ) 
XA(1, {                                           ) 
XB(2,   while (TRUE)                              ) 
XA(3,     {                                       ) 
XB(4,       P(x1);                                ) 
XP(5,       wjazd na stare miasto                 ) 
XB(6,       P(x5);                                )
XP(7,       przejazd przez 1                      ) 
XB(8,       V(x1);                                )
XB(9,       P(x3);                                ) 
XP(10,      przejazd przez 4                      ) 
XB(11,      V(x5);                                ) 
XB(12,      P(x2);                                ) 
XP(13,      przejazd przez 3                      ) 
XB(14,      V(x3);                                ) 
XP(15,      wyjazd ze starego miasta              ) 
XB(16,      V(x2);                                ) 
XB(17,    }                                       ) 
XB(18,}                                           ) 

XD(0, void busC(int i)                            ) 
XA(1, {                                           ) 
XB(2,   while (TRUE)                              ) 
XA(3,     {                                       ) 
XB(4,       P(x5);                                ) 
XB(5,       P(x3);                                ) 
XB(6,       P(x2);                                ) 
XP(7,       wjazd na stare miasto                 ) 
XP(8,       przejazd przez 3                      ) 
XB(9,       V(x2);                                )
XP(10,      przejazd przez 4                      ) 
XB(11,      V(x3);                                )
XP(12,      przejazd przez 1                      ) 
XB(13,      V(x5);                                ) 
XP(14,      przejazd przez 2                      ) 
XP(15,      wyjazd ze starego miasta              ) 
XB(16,    }                                       ) 
XB(17,}                                           ) 

XD(0, void busD(int i)                            ) 
XA(1, {                                           ) 
XB(2,   while (TRUE)                              ) 
XA(3,     {                                       ) 
XB(4,       P(x4);                                ) 
XP(5,       wjazd na stare miasto                 ) 
XB(6,       P(x1);                                ) 
XB(7,       P(x5);                                ) 
XP(8,       przejazd przez 4                      ) 
XB(9,       V(x4);                                )
XP(10,      przejazd przez 1                      ) 
XB(11,      V(x5);                                )
XP(12,      wyjazd ze starego miasta              ) 
XB(13,      V(x1);                                )
XB(14,    }                                       ) 
XB(15,}                                           ) 
void sopmpk_state()
{
  int v1,v2,v3,v4,v5;
  v1 = semctl(semid, x1, GETVAL, NULL);
  v2 = semctl(semid, x2, GETVAL, NULL);
  v3 = semctl(semid, x3, GETVAL, NULL);
  v4 = semctl(semid, x4, GETVAL, NULL);
  v5 = semctl(semid, x5, GETVAL, NULL);
  mvprintw(0, 0, "x1=%d x2=%d x3=%d x4=%d x5=%d", v1,v2,v3,v4,v5);
}

int main(int argc, char* argv[])
{
  int i, key;
  pthread_t* watki;

  if (argc<3)
    {
      printf("%s algorytm lprocesow\n\n"
	     "algorytmy:\talgorytm1, algorytm2, algorytm3, algorytm4, dekker, dijkstra, sopmpk\n"
	     "lprocesow:\tliczba procesow do uruchomienia (w niektorych algorytmach ignorowane)\n", argv[0]);
      // Tu nalezy dodac takze nazwe swojego algorytmu... Patrz tez nizej...
      exit(1);
    }

  n = atoi(argv[2]);

  if (n<=1)
    {
      printf("Musi byc conajmniej 2 procesy.\n");
      exit(1);
    }

  if (strcmp(argv[1], "algorytm1")==0)
    {
      n=2; // jedyna opcja, z jaka dzialamy w tym algorytmie...
      algorytm = &alg1_processi;
      state = &arg1_state;
    } 
  else if (strcmp(argv[1], "algorytm2")==0)
    {
      n=2;
      algorytm = &alg2_processi;
      state = &arg2_state;
    }
  else if (strcmp(argv[1], "algorytm3")==0)
    {
      n=2;
      algorytm = &alg3_processi;
      state = &arg3_state;
    }
  else if (strcmp(argv[1], "algorytm4")==0)
    {
      n=2;
      algorytm = &alg4_processi;
      state = &arg4_state;
    }
  else if (strcmp(argv[1], "dekker")==0)
    {
      n=2;
      algorytm = &dekker_processi;
      state = &dekker_state;
    }
  else if (strcmp(argv[1], "dijkstra")==0)
    {
      flag = (int*) malloc(sizeof(int)*n); 
      test = (int*) malloc(sizeof(int)*n); 
      k = (int*) malloc(sizeof(int)*n); 
      other = (int*) malloc(sizeof(int)*n); 
      bzero(flag, n*sizeof(int));
      bzero(test, n*sizeof(int));
      bzero(k, n*sizeof(int));
      bzero(other, n*sizeof(int));
      algorytm = &dijkstra_processi;
      state = &dijkstra_state;
    }
  else if (strcmp(argv[1], "sopmpk")==0)
    {
      semid = semget(IPC_PRIVATE, 5, IPC_CREAT|0644);
      if (semid==-1)
	{
	  perror("semget");
	  exit(1);
	}
      if (semctl(semid, x1, SETVAL, 1)==-1) { perror("semctl x1"); exit(1); }
      if (semctl(semid, x2, SETVAL, 1)==-1) { perror("semctl x1"); exit(1); }
      if (semctl(semid, x3, SETVAL, 1)==-1) { perror("semctl x1"); exit(1); }
      if (semctl(semid, x4, SETVAL, 1)==-1) { perror("semctl x1"); exit(1); }
      if (semctl(semid, x5, SETVAL, 1)==-1) { perror("semctl x1"); exit(1); }
      algorytmy[0] = &busA;
      algorytmy[1] = &busB;
      algorytmy[2] = &busC;
      algorytmy[3] = &busD;
      state = &sopmpk_state;
    }
  // Tu nalezy takze dodac przypadek obslugujacy inicjalizacje, jesli wybrany zostal przez uzytkownika
  // Twoj algorytm. A wiec przypisanie zmiennym algorytm i state odpowiednich funkcji, sprawdzenie czy
  // wartosc n jest dobra, oraz ewentualna inicjalizacja innych zmiennych.
      
  // inicjalizacja
  watki = (pthread_t*) malloc(sizeof(pthread_t)*n);
  m = (pthread_mutex_t*) malloc(sizeof(pthread_mutex_t)*n);
  initscr();
  cbreak();
  noecho();
  curs_set(0);
  getmaxyx(stdscr, h, w);

  mvprintw(h-1, 0, "0-9 - krok i-tego procesu      Ctrl+C - wyjscie");

  // odpalenie algorytmu
  state();
  for (i=0; i<n; i++)
    {
      pthread_mutex_init(&m[i], NULL);
      pthread_mutex_unlock(&m[i]);
      if (algorytmy[i]!=NULL)
	pthread_create(&watki[i], NULL, (void*)(algorytmy[i]), (void*)i);
      else
	pthread_create(&watki[i], NULL, (void*)algorytm, (void*)i);
    }

  while ((key = getch()) != 0)
    {
      key = key - '0';
      if (key>=0 && key<n)
	pthread_mutex_unlock(&m[key]);
      state();
    }

  for (i=0; i<n; i++)
    {
      pthread_join(watki[i], NULL);
    }

  endwin();
  return 0;
}
