import java.io.*;

/**
 *  Program sprawdza czy prosta opisana dwoma punktami przecina
 *  prostokt.
 * 
 *  Program czyta ze standardowego wejcia wsprzdne czterech punktw:
 *  Ax Ay Bx By Cx Cy Dx Dy
 *  Z ktrych punkty A i B nale do prostej, a punkty C i D stanowi
 *  dwa przeciwlege wierzchoki prostokta.
 *  Jeli prosta przecina prostokt, na standardowe wyjcie wypisywane
 *  jest TRUE, lub w przeciwnym wypadku FALSE.
 *
 *  @author Katarzyna Rafalska (inf66305), Jacek Pospychaa (inf66301)
 */
class inf66305_66301_io_lab9 {

    /**
     * Pole z ostatnio odczytan prost i prostoktem
     */
    static Dana d;

    /**
     *  W nieskoczonej ptli czyta dane i wywietla wynik.
     *  Wyjcie nastpuje w przypadku IOException
     */
    public static void main(String[] args)
	{
	    try {
		d = new Dana();
		while (true)
		    {
			d.czytaj();
			d.wyswietl_wynik();
		    }
		} catch (IOException e) {;}
	}
}
    
/**
 *  Punkt - podstawowy element, z ktrego skadaj si figury w programie
 *  Punkt jest opisany przez wsprzdne x i y.
 *
 *  @see Figura
 */
class Punkt
{
    double x;
    double y;

    /**
     *  Konstruktor, ustawiajcy punkt w pozycji opisanej wsprzdnymi
     *
     * @param x Nowa warto x
     * @param y Nowa warto y
     */
    Punkt(double x, double y)
    {
	this.x = x;
	this.y = y;
    }
    
    /**
     *  Konstruktor, ustawiajcy punkt w pozycji takiej samej, jak inny punkt
     *
     * @param p Punkt, ktrego wsprzdne maj by skopiowane
     */
    Punkt(Punkt p)
    {
	this.x = p.x;
	this.y = p.y;
    }
}

/**
 *  Abstrakcyjna klasa reprezentujca figury w programie. Kada figura skada si z punktw
 *  Poza tym, figury nie maj wicej wsplnych cech z punktu widzenia problemu.
 *
 *  @see Punkt
 */
abstract class Figura
{
    /**
     *  Punkty, z ktrych skada si figura
     */
    Punkt p[];
}

/**
 *  Prosta, opisywana dwoma punktami, ktre do niej nale. Dwa punkty
 *  jednoznacznie identyfikuj prost. W programie prosta jest figur.
 *
 *  @see Figura
 *  @see Punkt
 */
class Prosta extends Figura
{
    
    /**
     *  Konstruktor, tworzcy prost w oparciu o dwa punkty, nalece do niej
     *
     * @param a Pierwszy punkt nalecy do prostej
     * @param b Drugi punkt nalecy do prostej
     */
    Prosta(Punkt a, Punkt b)
    {
	this.p = new Punkt[2];
	this.p[0] = new Punkt(a);
	this.p[1] = new Punkt(b);
    }
    
    /**
     *  y(x) - Zwraca wsprzdn y punktu nalecego do prostej, 
     *         ktrego x podano w argumencie
     *
     *  Funkcja korzysta ze wzoru wyprowadzanego z oglnego rwnania prostej.
     *  Moe si wywali z bdem dzielenia przez zero, jeli prosta jest pionowa.
     *
     *  @param x Wsprzdna x poszukiwanego punktu
     *  @return Wsprzdna y poszukiwanego punktu
     *  @see Prosta#y
     */
    double y(double x)
    {
	return (p[1].y - p[0].y)/(p[1].x - p[0].x)*(x - p[0].x) + p[0].y;
    }
    
    /**
     *  x(y) - Zwraca wsprzdn x punktu nalecego do prostej, 
     *         ktrego y podano w argumencie
     *
     *  Funkcja korzysta ze wzoru wyprowadzanego z oglnego rwnania prostej.
     *  Moe si wywali z bdem dzielenia przez zero, jeli prosta jest pozioma.
     *
     *  @param y Wsprzdna y poszukiwanego punktu
     *  @return Wsprzdna x poszukiwanego punktu
     *  @see Prosta#x
     */
    double x(double y)
    {
	return (p[1].x - p[0].x)/(p[1].y - p[0].y)*(y - p[0].y) + p[0].x;
    }
    
}

/**
 *  Prostokt to figura. Opisywany jest punktami dwch skrajnych wierzchokw.
 *  Poniewa dla dwch skrajnych wierzchokw moe istnie wicej ni jeden
 *  prostokt, w klasie przyjto zaoenie, e przetwarzane s tylko prostokty
 *  o bokach rwnolegych/prostopadych do osi ukadu wsprzdnych.
 *
 *  @see Figura
 *  @see Punkt
 */
class Prostokat extends Figura
{
    
    /**
     *  Konstruktor, tworzcy prostokt o dwch skrajnych wierzchokach podanych w argumentach.     
     *
     *  Prostokt przed zapamitaniem jest uporzdkowywany tak, e pierwszy wierzchoek jest
     *  wierzchokiem najbliej ukadu wsprzdnych, a drugi punkt jest wierzchokiem najdalej
     *  ukadu wsprzdnych.
     *
     * @param a Pierwszy skrajny wierzchoek
     * @param b Drugi skrajny wierzchoek
     * @see Punkt
     */
    Prostokat(Punkt a, Punkt b)
    {
	this.p = new Punkt[2];
	this.p[0] = new Punkt(Math.min(a.x, b.x), Math.min(a.y, b.y));
	this.p[1] = new Punkt(Math.max(a.x, b.x), Math.max(a.y, b.y));
    }
    
    /**
     *  Funkcja, okrelajca czy ten prostokt jest przecinany przez prost podan w argumencie.
     *
     *  Funkcja sprawdza, czy ktry z bokw prostokta jest przecinany przez prost.
     *  Odcinek (bok prostokta) jest przecinany przez prost, jeli jego skrajne wierzchoki
     *  s po rnych stronach prostej, tzn. jeden jest ponad, a drugi pod prost.
     *  Korzystamy z uporzdkowania punktw prostokta nadanego podczas tworzenia obiektu.
     *
     *  @param pr Prosta, ktrej stan wzgldem prostokta sprawdzamy
     *  @see Prosta
     */
    boolean czy_nalezy(Prosta pr)
    {
	return ((p[1].y >= pr.y(p[0].x) && pr.y(p[0].x) >= p[0].y) ||
		(p[1].y >= pr.y(p[1].x) && pr.y(p[1].x) >= p[0].y) ||
		(p[1].x >= pr.x(p[0].y) && pr.x(p[0].y) >= p[0].x) ||
		(p[1].x >= pr.x(p[1].y) && pr.x(p[1].y) >= p[0].x));
    }
}

/**
 *  Gwny obiekt przetwarzania w programie, czyli prostokt i prosta, ktrych stosunek
 *  wzgldem siebie jest obliczany. Klasa zajmuje si pobieraniem danych ze standardowego
 *  wejcia, oraz wypisywaniem wynikw.
 *
 *  @see Prosta
 *  @see Prostokat
 */
class Dana
{
    Prosta prosta;
    Prostokat prostokat;
    
    /**
     *  Odczytuje ze standardowego wejcia jedn lini. Linia musi zawiera 8 liczb,
     *  oddzielonych spacjami:
     *  Ax Ay Bx By Cx Cy Dx Dy
     *  Z ktrych punkty A i B nale do prostej, a punkty C i D stanowi
     *  dwa przeciwlege wierzchoki prostokta. Aby wywietli wynik, naley korzysta
     *  z funkcji wyswietl_wynik.
     *
     *  @exception IOException rzuca wyjtek, gdy nie uda si odczyt
     *  @see Dana#wyswietl_wynik
     *  @see System.in
     */
    public void czytaj() throws IOException
    {
	String line;
	String s[];
	Double d[] = new Double[8];
	BufferedReader in = new BufferedReader(new InputStreamReader(System.in));

	if ((line = in.readLine()) != null)
	    {
		s = line.split(" ", 8);

		for (int i=0; i<s.length; i++)
		    d[i] = new Double(s[i]);

		prostokat = new Prostokat(new Punkt(d[0].doubleValue(), d[1].doubleValue()), 
					  new Punkt(d[2].doubleValue(), d[3].doubleValue()));		
		prosta = new Prosta(new Punkt(d[4].doubleValue(), d[5].doubleValue()), 
				    new Punkt(d[6].doubleValue(), d[7].doubleValue()));

	    }
    }
    
    /**
     *  Wywietla wynik dla danych prostej i prostokta w obiekcie. Jeli prosta przecina
     *  prostokt, na standardowe wyjcie wypisywane jest TRUE, w przeciwnym wypadku
     *  wypisywane jest FALSE. Aby wprowadzi dane, naley skorzysta z funkcji czytaj.
     *
     *  @see Dane#czytaj
     *  @see System.out
     */
    public void wyswietl_wynik()
    {
	if (prostokat.czy_nalezy(prosta))
	    System.out.println("TRUE");
	else
	    System.out.println("FALSE");
    }
}
