/****
   LTPS - Linux Text Presentation System
                                       ****/

#include <iostream>
#include <stdlib.h> // exit()
#include <getopt.h>

#include "options.h"

using namespace std;

extern int optind; // first non-option argument

Options::Options()
{
  this->filename = NULL;
}

/** Constructor which sets all command-line options
 *  Takes argc and argv from main() as an arguments
 *
 */
Options::Options(int argc, char** argv)
{
  int opt_code;

  while ((opt_code = getopt(argc, argv, "hv")) != -1)
    {
      switch (opt_code)
	{
	case 'h':
	  this->show_usage();
	  exit(1);
	  break;

	case 'v':
	  this->show_version();
	  exit(1);
	  break;
	}
    }

  if (argv[optind] != NULL) // last non-option arg is presentation filename
    {
      this->filename = argv[optind];
    }
  else // no filename given
    {
      cout << "Error: No filename given" << endl;
      this->show_usage();
      exit(1);
    }
}

char* Options::get_filename(void) { 
  return filename; 
} 


/* * * * * *
  PRIVATE:
 * * * * * */

/** Shows application command line arguments
 *  Also shows application version (see show_version)
 */
void Options::show_usage(void)
{
  this->show_version();
  cout << "usage: ltps filename" << endl
       << "\tfilename\txml file to load" << endl;
}

/** Shows application version
 */
void Options::show_version(void)
{
  cout << "Linux Text Presentation System (LTPS)" << endl;
}
