/*
  projekt na grafik komputerow

  Katarzyna Rafalska (66305)
  Jacek Pospychaa (66301)

*/

#include <GL/glut.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include <time.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include <sys/time.h>

#include "flipper_code.h"  

int window;
int mouse_button = 0;

float a = 0, b = 0, c = 0;
float camx = 0, camy = -7, camz = 1;
float camrx_factor = 0, camry_factor = 0, camrz_factor = 0;
float camrx = 0, camry = 45, camrz = 0;

int viewmode=1;

struct timeval left_hand_shoot = {0, 0};
struct timeval right_hand_shoot = {0, 0};

struct timeval last_idle_time;

list_t* ents = NULL;

list_t* walls = NULL; //list of planes

entity_t* left_hand = NULL;
entity_t* right_hand = NULL;
entity_t* ball = NULL;

plane_t* clhand = NULL;
plane_t* crhand = NULL;

vertex_t left_normal;
vertex_t right_normal;

vertex_t ball_start_pos;

float horiz_normal[3] = {0,0,1};

float ballfx = 0, ballfy = 0; // ball force vertex (x,y)

float tforce = 0.005;
float gravity = 0.002;

vertex_t* colors = NULL;
int* textures = NULL;
int colors_count;
int textures_count;

GLfloat viewer_light_ambient[] = {0.5f, 0.5f, 0.5f, 1.0f};
GLfloat viewer_light_diffuse[] = {1.0f, 0.0f, 0.0f, 1.0f};

GLfloat top_light_ambient[] = {0.3f, 0.3f, 0.3f, 1.0f};
GLfloat top_light_diffuse[] = {0.8f, 0.8f, 0.8f, 1.0f};
GLfloat top_light_position[] = {0.0f, 10.0f, 18.0f, 1.0f};

image_t cmap;// collisions_map
float ballx0=-1.23, bally0=-4;
float ballw=2.5, ballh=6;

#define BUFSIZE 1024

void free_polyline(polyline_t *p)
{
  int i;
  if (p!=NULL)
    {
      free(p->v);
      for (i=0; i<p->planes_count; i++)
	free(&(p->p[i]));
      free(p->p);
      free(p);
    }
}

void free_block(block_t *b)
{
  if (b!=NULL)
    {
      free(b->name);
      free_polyline(b->polyline);
      free(b);
    }
}

void free_entity(entity_t* e)
{
  if (e!=NULL)
    {
      free(e->name);
      free_block(e->b);
      free(e);
    }
}

polyline_t* copy_polyline(polyline_t *p)
{
  polyline_t *new_polyline = NULL;
  int i, j, k;
  vertex_t* vptr = NULL;

  if (p!=NULL)
    {
      new_polyline = (polyline_t*) malloc(sizeof(polyline_t));
      new_polyline->vertices_count = p->vertices_count;
      new_polyline->planes_count = p->planes_count;

      new_polyline->v = (vertex_t*) malloc(sizeof(vertex_t)*new_polyline->vertices_count);
      memcpy(new_polyline->v, p->v, sizeof(vertex_t)*new_polyline->vertices_count);
      
      new_polyline->p = (plane_t*) malloc(sizeof(plane_t)*new_polyline->planes_count);
      memcpy(new_polyline->p, p->p, sizeof(plane_t)*new_polyline->planes_count);
      for (i=0; i<new_polyline->planes_count; i++) // update pointers to vertices
	{
	  for (j=0; j<new_polyline->p[i].vertices_count; j++) // for every vertex in a plane
	    {
	      k=0;
	      vptr = NULL;
	      while ((k<new_polyline->vertices_count) && (vptr == NULL)) 
		{
		  if (&(p->v[k]) == p->p[i].v[j]) // search while vertices with the same pointers not found
		    vptr = &new_polyline->v[k];
		  else
		    k++;
		}
	      new_polyline->p[i].v[j] = vptr;
	    }
	}
      
      return new_polyline;
    }
  else
    return NULL;
}

block_t* copy_block(block_t *b)
{
  block_t *new_block = NULL;

  if (b!=NULL)
    {
      new_block  = (block_t*) malloc(sizeof(block_t));

      new_block->name = (char*) malloc(strlen(b->name)+1);
      bzero(new_block->name, strlen(b->name)+1);
      strcpy(new_block->name, b->name);

      new_block->polyline = copy_polyline(b->polyline);
      return new_block;
    }
  else
    return NULL;
}

void del_from_list(list_t **list)
{
  void* next;
  if (*list != NULL)
    {
      next = (*list)->next;
      free(*list);
      *list = next;
    }
}

void add_to_list(list_t** list, void* newitem)
{
  list_t* new = (list_t*) malloc(sizeof(list_t));
  bzero(new, sizeof(list_t));
  new->current = newitem;
  if (*list != NULL)
    {
      new->next = *list;
      *list = new;

    }
  else
    {
      *list = new;
    }
}

char* xfgets(char* buf, size_t size, FILE* stream, int *counter)
{
  (*counter)++;
  return fgets(buf, size, stream);
}

/* reads from stream, incrementing lines counter. expects group code 'code' (-1 -any), scans data
   by format (using scanf), and reads one item do dest.
   if dest is NULL, field is read, but not verified against format and skipped (useful for unwanted fields)
   on other code issues a warning message and returns -1. data is NOT read
   if scanf fails, issues a warning message and returns -2.
   normally returns read groupid.
*/
int readpaircodef(void* dest, char* format, int code, FILE* stream, int *counter)
{
  int groupid;
  char buf[BUFSIZE];
  xfgets(buf, BUFSIZE, stream, counter);
  if (sscanf(buf, "%d", &groupid) == 1)
    {
      if ((code!=-1) && (groupid != code))
	{
	  fseek(stream, -1*strlen(buf), SEEK_CUR);
	  return -1;
	}
      xfgets(buf, BUFSIZE, stream, counter);
      if (dest!=NULL)
	{
	  if (sscanf(buf, format, dest) != 1)
	    {
	      return -2;
	    }
	}
      return groupid;
    }
  else
    {
      return -1;
    }  
}

/* reads a group code and line after it. line is returned in buf, code is returned by result.
   on fail, returns -1.
*/
int readpair(char* buf, size_t size, FILE* stream, int *counter)
{
  return readpaircodef(buf, "%s", -1, stream, counter);
}


/* reads count vertices to v, not more, not less!! allocates memory by itself */
void readvertex(vertex_t* v, FILE* stream, int *counter)
{
  if (v!=NULL)
    {
      readpaircodef(&(v->x), "%f", DXF_CODE_X, stream, counter);
      readpaircodef(&(v->z), "%f", DXF_CODE_Y, stream, counter);
      v->z *= -1;
      readpaircodef(&(v->y), "%f", DXF_CODE_Z, stream, counter);
    }
}

void readvertices(vertex_t** v, int count, FILE* stream, int *counter)
{
  int i;

  *v = (vertex_t*) malloc(count*sizeof(vertex_t));
  //  printf("reading %d vertices", count);
  for (i=0; i<count; i++)
    {
      readpaircodef(NULL, "", DXF_CODE_ENTITY, stream, counter); // read VERTEX entity
      readpaircodef(NULL, "", DXF_CODE_LAYER, stream, counter); // ignore LAYER
      readvertex(&((*v)[i]), stream, counter);      
      readpaircodef(NULL, "", DXF_CODE_FLAGS, stream, counter); // ignore FLAGS
    }
  //  printf(".\n");
}

int intersect2d(plane_t* plane, float x, float y, float fx, float fy, float *d)
{
  double dotproduct;
  double l2;
  float mx, Mx, my, My;

  dotproduct  = fx*plane->normal.x + fy*plane->normal.z;
  if ((dotproduct < ZERO) && (dotproduct > -ZERO)) 
      return 0;

  l2 = (plane->normal.x*(plane->v[0]->x - x) + plane->normal.z*(plane->v[0]->z - y))/dotproduct;
  //  l2=(plane._Normal.dot(plane._Position-position))/dotproduct;

  if (l2 < -ZERO)
    return 0;
  
  mx = plane->v[0]->x < plane->v[1]->x ? plane->v[0]->x : plane->v[1]->x;
  my = plane->v[0]->z < plane->v[1]->z ? plane->v[0]->z : plane->v[1]->z;
  Mx = plane->v[0]->x > plane->v[1]->x ? plane->v[0]->x : plane->v[1]->x;
  My = plane->v[0]->z > plane->v[1]->z ? plane->v[0]->z : plane->v[1]->z;

  if ((l2<2) && (! ((mx<=x && x<=Mx) || (my<=y && y<=My))))
    return 0;

  *d = l2;
  return 1;
}

int BMPLoad(char *filename, image_t *image) 
{
    FILE *file;
    long size;
    int i;
    unsigned short int planes;
    unsigned short int bpp;
    char temp;         

    // make sure the file is there.
    if ((file = fopen(filename, "rb"))==NULL)
    {
	printf("File Not Found : %s\n",filename);
	return 0;
    }
    fseek(file, 18, SEEK_CUR);
    if ((i = fread(&image->sizeX, 4, 1, file)) != 1) {
	printf("Error reading width from %s.\n", filename);
	return 0;
    }
    if ((i = fread(&image->sizeY, 4, 1, file)) != 1) {
	printf("Error reading height from %s.\n", filename);
	return 0;
    }
    size = image->sizeX * image->sizeY * 3;
    if ((fread(&planes, 2, 1, file)) != 1) {
	printf("Error reading planes from %s.\n", filename);
	return 0;
    }
    if (planes != 1) {
	printf("Planes from %s is not 1: %u\n", filename, planes);
	return 0;
    }
    if ((i = fread(&bpp, 2, 1, file)) != 1) {
	printf("Error reading bpp from %s.\n", filename);
	return 0;
    }
    if (bpp != 24) {
	printf("Bpp from %s is not 24: %u\n", filename, bpp);
	return 0;
    }
    fseek(file, 24, SEEK_CUR);

    image->data = (char *) malloc(size);
    if (image->data == NULL) {
	printf("Error allocating memory for color-corrected image data");
	return 0;	
    }

    if ((i = fread(image->data, size, 1, file)) != 1) {
	printf("Error reading image data from %s.\n", filename);
	return 0;
    }

    for (i=0;i<size;i+=3) { // reverse all of the colors. (bgr -> rgb)
	temp = image->data[i];
	image->data[i] = image->data[i+2];
	image->data[i+2] = temp;
    }
    
    return 1;
}

int BMPcolor(image_t* image, int x, int y)
{
  if (x>(int)image->sizeX) 
    printf("cmap x overflow: %d (max %d)\n", x, (int)image->sizeX); 
  else if (x<0) 
    printf("cmap x underflow: %d\n", x); 
  else if (y>(int)image->sizeY) 
    printf("cmap y overflow: %d (max %d)\n", y, (int)image->sizeY); 
  else if (y<0) 
    printf("cmap y overflow: %d\n", y); 
  else return image->data[(y*image->sizeX + x)*3];

  return 0;
}

int bind_textures(char *filename)
{
  FILE* fp;
  int i, j;
  char buf[BUFSIZE];
  int count, plane_num, tex_num;
  list_t* item;
  entity_t* e;


  printf("Binding textures\n");
  if ((fp = fopen(filename, "r")) == NULL)
    {
      perror("can't open file");
      return -1;
    }

  fscanf(fp, "%d", &count);

  for (i=0; i<count; i++)
  {
      fscanf(fp, "%s %d %d", buf, &plane_num, &tex_num);

      item = ents;
      while (item != NULL)
      {
	  e = (entity_t*)(item->current);
	  if (strcmp(buf, e->name) == 0)
	    {
	      printf("texture %d to %s plane%d\n", tex_num, buf, plane_num);
	      //	      e->b->polyline->p[plane_num].color = 0;
	      e->b->polyline->p[plane_num].texnum = tex_num;
	      
	      for (j=0; j< e->b->polyline->p[plane_num].vertices_count; j++)
		{
		  fscanf(fp, "%f %f",
			 &e->b->polyline->p[plane_num].coords[j][0],
			 &e->b->polyline->p[plane_num].coords[j][1]
			 );
		}
	    }
	  item = item->next;
      }
  }

  fclose(fp);
  return 0;
}

int read_colors(char *filename)
{
  FILE* fp;
  image_t image;
  int i;
  char buf[BUFSIZE];
  printf("Loading colors\n");

  if ((fp = fopen(filename, "r")) == NULL)
    {
      perror("can't open file");
      return -1;
    }

  fscanf(fp, "%d %d", &colors_count, &textures_count);

  colors = (vertex_t*) malloc(colors_count*sizeof(vertex_t));
  textures = (int*) malloc(textures_count*sizeof(int));
  glGenTextures(textures_count, textures);

  for (i=0; i<colors_count; i++)
    {
      fscanf(fp, "%f %f %f", &(colors[i].x), &(colors[i].y), &(colors[i].z));
    }

  printf("Loading textures\n");
  for (i=0; i<textures_count; i++)
    {
      fscanf(fp, "%s", buf);
      if (!BMPLoad(buf, &image))
	{
	  printf("cant open image %s\n", buf);
	  exit(1);
	}

      glBindTexture(GL_TEXTURE_2D, textures[i]);
      glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR); // scale linearly when image bigger than texture
      glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR); // scale linearly when image smaller than texture
      glTexImage2D(GL_TEXTURE_2D, 0, 3, image.sizeX, image.sizeY, 0, GL_RGB, GL_UNSIGNED_BYTE, image.data);
    }
  fclose(fp);
  return 0;
}

/* returns list of entity_t with complete objects */
list_t* read_dxf(char *filename)
{
  FILE *fp;
  int groupid = 0;
  int lnum=0;
  char buf[BUFSIZE];
  int tmpi, tmpii;
  float tmpf;
  int ignore=0;

  polyline_t *current_polyline = NULL;
  block_t *new_block = NULL, *block = NULL;
  entity_t* new_entity = NULL;

  list_t *blocks_list = NULL, *ents_list = NULL, *list_item = NULL;

  if ((fp = fopen(filename, "r")) == NULL)
    {
      perror("can't open file");
      return NULL;
    }

  while ((groupid >= 0) && (groupid = readpair(buf, BUFSIZE, fp, &lnum)) >= 0)
    {
      tmpf=-1;
      tmpi=-1;
      tmpii=-1;

      if ((! ignore) || (groupid == 0))
	{
	  switch (groupid)
	    {
	    case (0): // entity
	      if (strncmp(buf, "SECTION", 7) == 0) // entity is a section
		{	
		  if  (readpair(buf, BUFSIZE, fp, &lnum) == 2)
		    {
		      //		      printf("Section: %s\n", buf);
		      if (strncmp(buf, "HEADER", 6) == 0) // ignore everyting in section HEADER
			{
			  printf("ignoring everything\n");
			  ignore = 1;
			}
		      else if (strncmp(buf, "BLOCKS", 6) == 0) // entity is blocks
			{
			  printf("Blocks\n");
			  ignore = 0;
			}
		      else if (strncmp(buf, "ENTITIES", 8) == 0) // entities
			{
			  printf("Entities\n");
			  ignore = 0;
			}
		      else
			{
			  printf("Unknown section, ignoring\n");
			  ignore = 1;
			}
		    }
		  else
		    {
		      printf("enexpected something after section: %s at line %d\n", buf, lnum);
		    }
		}
	      else if (strncmp(buf, "ENDSEC", 6) == 0) // entity is endsec
		{
		  //		  printf("End of section\n\n");
		}
	      else if (strncmp(buf, "BLOCK", 5) == 0) // entity is block
		{
		  //		  printf("Block\n");
		  new_block = (block_t*) malloc(sizeof(block_t));
		  bzero(new_block, sizeof(block_t));
		  new_block->polyline = (polyline_t*) malloc(sizeof(polyline_t));		  
		  bzero(new_block->polyline, sizeof(polyline_t));
		  new_block->name = (char*) malloc(BUFSIZE);
		  readpaircodef(new_block->name, "%s", DXF_CODE_NAME, fp, &lnum);
		  readpaircodef(NULL, "", DXF_CODE_LAYER, fp,&lnum);
		  readpaircodef(NULL, "", DXF_CODE_FLAGS, fp,&lnum);
		  readvertex(&(new_block->base), fp, &lnum);
		  current_polyline = new_block->polyline;

		  //  printf("base: %f %f %f\n", new_block->base.x, new_block->base.y, new_block->base.z);
		  add_to_list(&blocks_list, new_block);
		}
	      else if (strncmp(buf, "POLYLINE", 8) == 0) // polyline
		{
		  //		  printf("Polyline\n");
		  if (current_polyline != NULL)
		    {
		      readpaircodef(NULL, "", DXF_CODE_ENTITIES_FOLLOW, fp, &lnum);
		      readpaircodef(NULL, "", DXF_CODE_LAYER, fp,&lnum);
		      readpaircodef(&(current_polyline->color), "%d", DXF_CODE_COLOR, fp, &lnum);
		      readpaircodef(NULL, "", DXF_CODE_FLAGS, fp,&lnum);
		      current_polyline->vertices_count = 10;
		      readpaircodef(&(current_polyline->vertices_count), "%d", 71, fp, &lnum);
		      readpaircodef(&(current_polyline->planes_count), "%d", 72, fp, &lnum);
		      readvertices(&(current_polyline->v), current_polyline->vertices_count, fp, &lnum);

		      current_polyline->p = (plane_t*) malloc(sizeof(plane_t)*current_polyline->planes_count);
		      
		      for (tmpi = 0; tmpi<current_polyline->planes_count; tmpi++) // tmpi - current plane index
			{
			  readpaircodef(NULL, "", DXF_CODE_ENTITY, fp,&lnum); // read VERTEX
			  readpaircodef(NULL, "", DXF_CODE_LAYER, fp,&lnum); // read LAYER
			  readpaircodef(&(current_polyline->p[tmpi].color), "%d", DXF_CODE_COLOR, fp, &lnum);
			  readvertex(&(current_polyline->p[tmpi].normal), fp, &lnum);
			  printf("normal: %f %f %f\n", current_polyline->p[tmpi].normal.x, current_polyline->p[tmpi].normal.y, current_polyline->p[tmpi].normal.z);
			  readpaircodef(NULL, "", DXF_CODE_FLAGS, fp,&lnum); // read FLAGS

			  readpaircodef(&tmpii, "%d", 71, fp, &lnum);
			  current_polyline->p[tmpi].v[0] = &(current_polyline->v[tmpii-1]);
			  
			  readpaircodef(&tmpii, "%d", 72, fp, &lnum);
			  current_polyline->p[tmpi].v[1] = &(current_polyline->v[tmpii-1]);
			  
			  readpaircodef(&tmpii, "%d", 73, fp, &lnum);
			  current_polyline->p[tmpi].v[2] = &(current_polyline->v[tmpii-1]);

			  if (readpaircodef(&tmpii, "%d", -1, fp, &lnum) == 74)
			    {
			      current_polyline->p[tmpi].vertices_count = 4;
			      current_polyline->p[tmpi].v[3] = &(current_polyline->v[tmpii-1]);
			    }
			  else
			    {
			      current_polyline->p[tmpi].vertices_count = 3;
			    }

			  current_polyline->p[tmpi].normal.x = (current_polyline->p[tmpi].v[1]->y - current_polyline->p[tmpi].v[0]->y)*
                                                               (current_polyline->p[tmpi].v[2]->z - current_polyline->p[tmpi].v[1]->z) -
 			                                       (current_polyline->p[tmpi].v[1]->z - current_polyline->p[tmpi].v[0]->z)*
 			                                       (current_polyline->p[tmpi].v[2]->y - current_polyline->p[tmpi].v[1]->y);

			  current_polyline->p[tmpi].normal.y = (current_polyline->p[tmpi].v[1]->x - current_polyline->p[tmpi].v[0]->x)*
                                                               (current_polyline->p[tmpi].v[2]->z - current_polyline->p[tmpi].v[1]->z) -
 			                                       (current_polyline->p[tmpi].v[1]->z - current_polyline->p[tmpi].v[0]->z)*
 			                                       (current_polyline->p[tmpi].v[2]->x - current_polyline->p[tmpi].v[1]->x);

			  current_polyline->p[tmpi].normal.z = (current_polyline->p[tmpi].v[1]->x - current_polyline->p[tmpi].v[0]->x)*
                                                               (current_polyline->p[tmpi].v[2]->y - current_polyline->p[tmpi].v[1]->y) -
 			                                       (current_polyline->p[tmpi].v[1]->y - current_polyline->p[tmpi].v[0]->y)*
 			                                       (current_polyline->p[tmpi].v[2]->x - current_polyline->p[tmpi].v[1]->x);
			  
			  if (tmpi==1 || tmpi==2 || tmpi==3 || tmpi==5 || tmpi==0)
			    {
			      current_polyline->p[tmpi].normal.x *= -1;
			      current_polyline->p[tmpi].normal.y *= -1;
			      current_polyline->p[tmpi].normal.z *= -1;
			    }

			  // normalize
			  float d = sqrt(current_polyline->p[tmpi].normal.x*current_polyline->p[tmpi].normal.x +
					 current_polyline->p[tmpi].normal.y*current_polyline->p[tmpi].normal.y +
					 current_polyline->p[tmpi].normal.z*current_polyline->p[tmpi].normal.z);
			  current_polyline->p[tmpi].normal.x /= d;
			  current_polyline->p[tmpi].normal.y /= d;
			  current_polyline->p[tmpi].normal.z /= d;
			}

		      current_polyline = NULL; // avoid overwrite of new data by forgetting pointer to polyline
		      }
		} // end of polyline
	      else if (strncmp(buf, "ENDBLK", 6) == 0) // end of block
		{
		  //		  printf("End of block\n\n");
		}
	      else if (strncmp(buf, "VERTEX", 5) == 0) // vertex
		{
		  printf("unexpected vertex\n");
		}
	      else if (strncmp(buf, "SEQEND", 6) == 0) // seqend
		{		  //		  printf("Seqend\n\n");
		}
	      else if (strncmp(buf, "INSERT", 6) == 0) // insert entity
		{
		  new_entity = (entity_t*) malloc(sizeof(entity_t));
		  bzero(new_entity, sizeof(entity_t));
		  //		  printf("Entity\n");
		  readpaircodef(NULL, "", DXF_CODE_LAYER, fp,&lnum); // ignore LAYER
		  new_entity->name = (char*)malloc(BUFSIZE);
		  readpaircodef(new_entity->name, "%s", DXF_CODE_NAME, fp,&lnum);
		  readvertex(&(new_entity->base), fp,&lnum);
		  readpaircodef(&(new_entity->scale.x), "%f", DXF_CODE_XSCALE, fp,&lnum);
		  readpaircodef(&(new_entity->scale.z), "%f", DXF_CODE_YSCALE, fp,&lnum);
		  readpaircodef(&(new_entity->scale.y), "%f", DXF_CODE_ZSCALE, fp,&lnum);
		  readpaircodef(&(new_entity->text_rotation), "%f", DXF_CODE_TEXTROTATION, fp,&lnum);
		  printf("Shape %s: base x=%f y=%f z=%f\n", 
			 new_entity->name, 
			 new_entity->base.x, new_entity->base.y, new_entity->base.z);

		  list_item = blocks_list;
		  while ((list_item != NULL) && (new_entity->b == NULL))
		    {
		      block = list_item->current;
		      //		      printf("step: %s <=> %s\n", new_entity->name, block->name);
		      if (strcmp(new_entity->name, block->name) == 0)
			{
			  //			  printf("copying block %s\n", new_entity->name);
			  new_entity->b = copy_block(block);
			}
		      list_item = list_item->next;
		    }

		  if (strcmp(new_entity->name, "left_hand") == 0)
		    {
		      printf("left hand found\n");
		      left_hand = new_entity;
		    }
		  else if (strcmp(new_entity->name, "right_hand") == 0)
		    {
		      printf("right hand found\n");
		      right_hand = new_entity;
		    }
		  else if (strcmp(new_entity->name, "ball") == 0)
		    {
		      printf("ball found\n");
		      ball = new_entity;
		      memcpy(&ball_start_pos, &ball->base, sizeof(vertex_t));
		    }		  

		  add_to_list(&ents_list, new_entity);
		}
	      else if (strncmp(buf, "EOF", 3) == 0) // end of file
		{
		  printf("End\n");
		  groupid = -1;
		}
	      else // unknown entity
		{
		  printf("Unknown entity: %s\n", buf);
		}
	      break; // end of entity
	    case (2):
	      printf("IGNORED 2Name: %s\n", buf);
	      break;
	    case (3):
	      //	      printf("IGNORED 3Name: %s\n", buf);
	      break;
	    case (8): // layer
	      printf("IGNORED Layer %s\n", buf);
	      break;
	    case (10): // x coord
	      printf("unexpected x coord\n");
	      break;
	    case (20): // y coord
	      printf("unexpected y coord\n");
	      break;
	    case (30): // z coord
	      printf("unexpected z coord\n");
	      break;
	    case (41):
	      sscanf(buf, "%f", &tmpf);
	      printf("X scale factor %f\n", tmpf);
	      break;
	    case (42):
	      sscanf(buf, "%f", &tmpf);
	      printf("Y scale factor %f\n", tmpf);
	      break;
	    case (43):
	      sscanf(buf, "%f", &tmpf);
	      printf("Z scale factor %f\n", tmpf);
	      break;
	    case (50):
	      sscanf(buf, "%f", &tmpf);
	      printf("Text rotation %f\n", tmpf);
	      break;	      
	    case (62):
	      printf("Color number %s", buf);
	      break;
	    case (66):
	      printf("Entities follow\n");
	      break;
	    case (70):
	      sscanf(buf, "%d", &tmpi);
	      printf("IGNORED flags\n");
	      break;
	    case (71):
	      printf("Polygon mesh m=%s\n", buf);
	      break;
	    case (72):
	      printf("Polygon mesh n=%s\n", buf);
	      break;
	    case (73):
	      printf("Smooth surface density m=%s\n", buf);
	      break;
	    case (74):
	      printf("Smooth surface density n=%s", buf);
	      break;
	    case (999): // comment
	      xfgets(buf, BUFSIZE, fp, &lnum);
	      printf("Comment: %s\n", buf);
	      break;
	    default:
	      printf("Unknown group id \"%d\" content: \"%s\" at line %d\n", groupid, buf, lnum);
	    }
	}
    }

  fclose(fp);

  if ((left_hand == NULL) || (right_hand == NULL) || (ball == NULL))
    {
      printf("left hand, right hand or ball object not found\n");
      exit(1);
    }

  return ents_list;
}

void IdleStuff()
{
  struct timeval time_now;
  float dt, tmp, dist;
  int collision=0;
  list_t* item;
  plane_t* pt;

  gettimeofday(&time_now, NULL);

  dt = (float)(time_now.tv_sec  - last_idle_time.tv_sec) +  1.0e-6*(time_now.tv_usec - last_idle_time.tv_usec);
  ballfy += gravity;

  //  printf("check ball (%f,%f,%f)\n", ball->base.x, ball->base.y, ball->base.z);
  item = walls;
  while ((item!=NULL) && (collision==0))
    {
      pt = (plane_t*)item->current;
      collision = intersect2d(pt, ball->base.x, ball->base.z, ballfx, ballfy, &dist);
      if (dist>2) collision=0;
      item = item->next;
    }

  if (collision)
    { 
      top_light_diffuse[2] = 0.7;
      top_light_diffuse[0] = top_light_diffuse[1] = 1;

      tmp = 2* (-ballfx*pt->normal.x - ballfy*pt->normal.z);

      ballfx += tmp*pt->normal.x;
      ballfy += tmp*pt->normal.z;

      if (((pt == clhand) && (left_hand->rotation.y!=0)) || ((pt == crhand && (right_hand->rotation.y!=0))))
	{
	  printf("hand\n");
	  ballfx *= 4;
	  ballfy *= 4;
	}
      if (strcmp(pt->name, "wall_001") == 0 && dist<=1)
	{
	  printf("ball lost!\n");
	  memcpy(&ball->base, &ball_start_pos, sizeof(vertex_t));
	  ballfx = 0;
	  ballfy = 0;
	  sleep(1);
	}
      else
	{
	  if (rand()>RAND_MAX/2)
	    {
	      ballfx *= 1.05;
	      ballfy *= 1.05;
	    }
	  else
	    {
	      // loose some foce
	      ballfx *= 0.7;
	      ballfy *= 0.7;
	    }
	}
    }

  if (! collision)
    {
      top_light_diffuse[0] = top_light_diffuse[1] = top_light_diffuse[2] = 0.8;
      
      ball->base.x += ballfx;
      ball->base.z += ballfy;     
    }

  camrx += 2* camrx_factor * dt;
  camry += 2* camry_factor * dt;
  camrz += 2* camrz_factor * dt;

  if (left_hand_shoot.tv_usec != 0)
    {
      if ((float)(time_now.tv_sec  - left_hand_shoot.tv_sec) +  1.0e-6*(time_now.tv_usec - left_hand_shoot.tv_usec) > 0.5)
	{
	  memcpy(&clhand->normal, &left_normal, sizeof(vertex_t));
	  left_hand->rotation.y = 0;
	  bzero(&left_hand_shoot, sizeof(struct timeval));
	}
      else
	left_hand->rotation.y = 30;
    }

  if (right_hand_shoot.tv_usec != 0)
    {
      if ((float)(time_now.tv_sec  - right_hand_shoot.tv_sec) +  1.0e-6*(time_now.tv_usec - right_hand_shoot.tv_usec) > 0.5)
	{
	  memcpy(&crhand->normal, &right_normal, sizeof(vertex_t));
	  right_hand->rotation.y = 0;
	  bzero(&right_hand_shoot, sizeof(struct timeval));
	}
      else
	right_hand->rotation.y = -30;
    }

  last_idle_time = time_now;

  glutPostRedisplay();
}

list_t* extract_walls()
{
  int i, j, k=0;
  list_t *item, *planes = NULL;
  entity_t* e;
  polyline_t *p;
  vertex_t a, b, c, *tmp;
  plane_t* pt;

  item = ents;
  while (item != NULL)
    {
      e = (entity_t*)(item->current);
      p = e->b->polyline;
      if ((strncmp("wall", e->name, 4) == 0) || (strncmp("left_hand", e->name, 9) == 0) || (strncmp("right_hand", e->name, 9) == 0))
	{
	  for (i=0; i<p->planes_count; i++)
	    {
	      for (j=0; j<p->p[i].vertices_count; j++)
		{
		  memcpy(&a, p->p[i].v[j], sizeof(vertex_t));
		  memcpy(&b, p->p[i].v[(j+1)%(p->p[i].vertices_count)], sizeof(vertex_t));
		  memcpy(&c, p->p[i].v[(j+2)%(p->p[i].vertices_count)], sizeof(vertex_t));

		  a.x += e->base.x; a.y += e->base.y; a.z += e->base.z;
		  b.x += e->base.x; b.y += e->base.y; b.z += e->base.z;
		  c.x += e->base.x; c.y += e->base.y; c.z += e->base.z;

		  // if vertices a and b are at the same altitude and next vertex is higher (this means, plane is not laying on board)
		  // and it is lower than ball center
		  if ((fabs(a.y - b.y) < 0.01) && (c.y > ball->base.y) && (a.y < ball->base.y))
		    {
		      pt = (plane_t*) malloc(sizeof(plane_t));
		      memcpy(pt, &(p->p[i]), sizeof(plane_t));

		      tmp = (vertex_t*) malloc(sizeof(vertex_t)*2);
		      pt->v[0] = &tmp[0];
		      pt->v[1] = &tmp[1];

		      pt->name = e->name;
		      pt->vertices_count = 2;
		      
		      memcpy(pt->v[0], &a, sizeof(vertex_t));
		      memcpy(pt->v[1], &b, sizeof(vertex_t));
		      
		      if (strncmp("left_hand", e->name, 9) == 0)
			{
			  clhand = pt;
			  memcpy(&left_normal, &pt->normal, sizeof(vertex_t));
			}
		      if (strncmp("right_hand", e->name, 9) == 0)
			{
			  crhand = pt;
			  memcpy(&right_normal, &pt->normal, sizeof(vertex_t));

			  memcpy(pt->v[0], &b, sizeof(vertex_t));
			  memcpy(pt->v[1], &a, sizeof(vertex_t));
			}

		      printf("%d wall found (%f,%f <=> %f,%f) (%s)\n", ++k, 
			     pt->v[0]->x, 
			     pt->v[0]->z, 
			     pt->v[1]->x,
			     pt->v[1]->z,
			     e->name);
		      add_to_list(&planes, pt);
		      break;
		    }
		}
	    }
	}
      item = item->next;
    }
  return planes;
}

/* The main drawing function. */
void DrawGLScene()
{
  int i, j;
  list_t *item;
  entity_t* e;
  polyline_t *p;
  GLfloat light_position[] = {camx, 1+camy, 6+camz, 1.0f};
  GLfloat light_exponent[] = {100, 100, 100};
  GLfloat spot_dir[] = {ball->base.x, ball->base.y, ball->base.z};
  GLfloat light_cutoff[] = {10, 10, 10};

  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);		// Clear The Screen And The Depth Buffer
  glLoadIdentity();				// Reset The View

  glLightfv(GL_LIGHT2, GL_DIFFUSE, top_light_diffuse);

  glLightfv(GL_LIGHT1, GL_POSITION, light_position);
  glLightfv(GL_LIGHT1, GL_SPOT_DIRECTION, spot_dir);
  glLightfv(GL_LIGHT1, GL_SPOT_EXPONENT, light_exponent);
  glLightfv(GL_LIGHT1, GL_SPOT_CUTOFF, light_cutoff);


  switch (viewmode)
    {
    case 2:
      glRotatef(camrx, 0, -3, 0);
      glRotatef(camry, 3, 0, 0);
      glRotatef(camrz, 0, 0, -3);
      
      glTranslatef(camx, camy-1, camz+1);
      break;
    case 1:
      gluLookAt(camx, camy+11, camz-4, ball->base.x/10, ball->base.y, ball->base.z, 0, 1, 0);
      break;
    }

  //  glColor3f(1.0f,1.0f,1.0f);

  // send all objects to opengl
  item = ents;
  while (item != NULL)
    {
      e = (entity_t*)(item->current);
      p = e->b->polyline;

      glPushMatrix();

      glTranslatef(e->base.x, e->base.y, e->base.z);
      glTranslatef(e->b->base.x, e->b->base.y, e->b->base.z);

      glRotatef(e->rotation.y, 0, 1.0f, 0);
      glRotatef(e->rotation.x, 1.0f, 0, 0);
      glRotatef(e->rotation.z, 0, 0, 1.0f);

      for (i=0; i<p->planes_count; i++)
	{
	  

	  if (p->p[i].texnum != 0)
	    { // wersja texturowa
	      glBindTexture(GL_TEXTURE_2D, textures[p->p[i].texnum]);
	    }
	  else
	    glBindTexture(GL_TEXTURE_2D, textures[8]);	    

	      if (p->p[i].vertices_count==3)
		{
		  glBegin(GL_TRIANGLES);
		}
	      else
		{
		  glBegin(GL_QUADS);
		}

	      //glNormal3f(0, 0, 0);
	      glNormal3f(p->p[i].normal.x, p->p[i].normal.y, p->p[i].normal.z);
	      for (j=0; j<p->p[i].vertices_count; j++)
		{
		  glTexCoord2f(p->p[i].coords[j][0], p->p[i].coords[j][1]);
		  glVertex3f(p->p[i].v[j]->x, p->p[i].v[j]->y, p->p[i].v[j]->z);
		}
	      
	      /*	    }
	  else
	    { // wersja colorowa
	      glColor3f(1, 1, 1);
	      if (p->p[i].vertices_count==3) 
		glBegin(GL_TRIANGLES);
	      else 
		glBegin(GL_QUADS);	      

	      //	      glNormal3f(p->p[i].normal.x, p->p[i].normal.y, p->p[i].normal.z);
	      for (j=0; j<p->p[i].vertices_count; j++)
		glVertex3f(p->p[i].v[j]->x, p->p[i].v[j]->y, p->p[i].v[j]->z);
		}
	      */
	  glEnd();
	}

      glPopMatrix();

      item = item->next;
    }

  // since this is double buffered, swap the buffers to display what just got drawn.
  glutSwapBuffers();
}

/* A general OpenGL initialization function.  Sets all of the initial parameters. */
void InitGL(int Width, int Height)	        // We call this right after our OpenGL window is created.
{
  if (read_colors("colors") == -1)
    {
      printf("Failed to open file \"colors\"\n");
      exit(1);
    }

  srand(time(NULL));

  glClearColor(0.1f, 0.1f, 0.1f, 1.0f);		// This Will Clear The Background Color To Black
  glClearDepth(1.0);				// Enables Clearing Of The Depth Buffer
  glDepthFunc(GL_LESS);				// The Type Of Depth Test To Do
  glEnable(GL_DEPTH_TEST);			// Enables Depth Testin
//  glEnable(GL_TEXTURE_2D);
  glShadeModel(GL_SMOOTH);			// Enables Smooth Color Shading
  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);			// Really Nice Perspective Calculations
  glEnable(GL_LIGHTING);

  BMPLoad("textures/collisions.bmp", &cmap);

  bind_textures("tex_bindings");
  walls = extract_walls();

  glLightfv(GL_LIGHT1, GL_AMBIENT, viewer_light_ambient);
  glLightfv(GL_LIGHT1, GL_DIFFUSE, viewer_light_diffuse);
  glEnable(GL_LIGHT1);

  glLightfv(GL_LIGHT2, GL_AMBIENT, top_light_ambient);
  glLightfv(GL_LIGHT2, GL_POSITION, top_light_position);
  glEnable(GL_LIGHT2);

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();				// Reset The Projection Matrix

  gluPerspective(45.0f,(GLfloat)Width/(GLfloat)Height,0.1f,100.0f);	// Calculate The Aspect Ratio Of The Window

  glMatrixMode(GL_MODELVIEW);
}

/* The function called when our window is resized (which shouldn't happen, because we're fullscreen) */
void ReSizeGLScene(int Width, int Height)
{
  if (Height==0)				// Prevent A Divide By Zero If The Window Is Too Small
    Height=1;

  glViewport(0, 0, Width, Height);		// Reset The Current Viewport And Perspective Transformation

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(45.0f,(GLfloat)Width/(GLfloat)Height,0.1f,100.0f);
  glMatrixMode(GL_MODELVIEW);
}

void mouseMotion(int x, int y)
{
  //  static int last_x = -1;
  //  static int last_y = -1;

  camrx += (x/600)*60 - 30;

  printf("%d %d\n", x, y);
  glutPostRedisplay();

  /*
  int dx, dy;

  if (last_x == -1)
    {
      last_x = x;
      last_y = y;
    }

  if (mouse_button != 0)
    {
      dx = x - last_x;
      dy = y - last_y;
    }

  switch (mouse_button)
    {
    case GLUT_LEFT_BUTTON:
      camy += dy/10;
      camx += dx/10;
      break;
    case GLUT_RIGHT_BUTTON:
      camry += dy/10;
      camrx += dx/10;
    }

  last_x = x;
  last_y = y;  
  */
}

void mouseProc(int button, int state, int x, int y)
{
  mouse_button = button;
  glutPostRedisplay();
}

/* The function called whenever a key is pressed. */
void keyPressed(unsigned char key, int x, int y) 
{
  /* If escape is pressed, kill everything. */
  switch (key)
    {
      // translate
    case 'w':
      camy -= 0.3;
      break;
    case 's':
      camy += 0.3;
      break;
    case 'a':
      camx += 0.3;
      break;
    case 'd':
      camx -= 0.3;
      break;
    case 'q':
      camz += 0.3;
      break;
    case 'e':
      camz -= 0.3;
      break;
      // rotate
    case 't':
      camry_factor -= 1;
      break;
    case 'g':
      camry_factor += 1;
      break;
    case 'f':
      camrx_factor += 1;
      break;
    case 'h':
      camrx_factor -= 1;
      break;
    case 'r':
      camrz_factor += 1;
      break;
    case 'y':
      camrz_factor -= 1;
      break;


    case 'i':
      //      ballfy -= 0.02;
      b += 0.3;
      break;
    case 'k':
      // ballfy += 0.02;
      b -= 0.3;
      break;
    case 'j':
      // ballfx -= 0.02;
      a -= 0.3;
      break;
    case 'l':      
      // ballfx += 0.02;
      a += 0.3;
      break;
    case 'u':
      c += 0.3;
      break;
    case 'o':
      c += 0.3;
      break;
    case 'z':
      ballfx = ball->base.x;
      ballfy = ball->base.z;
      break;
    case '1':
      viewmode=1;
      break;
    case '2':
      viewmode=2;
      break;

    case 'c': // left hand
      {
	memcpy(&clhand->normal, horiz_normal, sizeof(vertex_t));
	left_hand->rotation.y = 30;
	gettimeofday(&left_hand_shoot, NULL);
	break;
      }
    case 'm': // right hand
      {
	memcpy(&crhand->normal, horiz_normal, sizeof(vertex_t));
	right_hand->rotation.y = -30;
	gettimeofday(&right_hand_shoot, NULL);
	break;
      }
    case ' ': // ball shoot

      break;
      // function keys
    case ESCAPE:
	glutDestroyWindow(window); 
	exit(0);
    }

  //    printf("ball (%f,%f) => (%f,%f)\n", ball->base.x, ball->base.z, ballfx, ballfy);
  glutPostRedisplay();
}

