/*
 projekt na grafik komputerow

  Katarzyna Rafalska (66305)
  Jacek Pospychaa (66301)

*/

#ifndef FLIPPER_CODE_H
#define FLIPPER_CODE_H

#define DXF_CODE_ENTITY          0
#define DXF_CODE_NAME            2
#define DXF_CODE_LAYER           8
#define DXF_CODE_X               10
#define DXF_CODE_Y               20
#define DXF_CODE_Z               30
#define DXF_CODE_XSCALE          41
#define DXF_CODE_YSCALE          42
#define DXF_CODE_ZSCALE          43
#define DXF_CODE_TEXTROTATION    50
#define DXF_CODE_COLOR           62
#define DXF_CODE_ENTITIES_FOLLOW 66
#define DXF_CODE_FLAGS           70

#define EPSILON 1.0e-8
#define ZERO EPSILON

typedef struct list_t
{
  void* current;
  struct list_t* next;
} list_t;

typedef struct vertex_t
{
  float x, y, z;
} vertex_t;

typedef struct plane_t {
  char* name;
  int vertices_count;
  int color;
  int texnum;
  float coords[4][2];
  vertex_t normal;
  vertex_t* v[4];
} plane_t;

typedef struct polyline_t {
  int vertices_count;
  int planes_count;
  int color;
  vertex_t* v;
  plane_t* p;
} polyline_t;

typedef struct block_t
{
  vertex_t base;
  char* name;
  polyline_t* polyline;
} block_t;

typedef struct entity_t
{
  char* name;
  block_t *b;
  vertex_t base;
  vertex_t scale;
  vertex_t rotation;
  float text_rotation;
    
} entity_t;

typedef struct image_t {
    unsigned long sizeX;
    unsigned long sizeY;
    char *data;
} image_t;

void InitGL(int Width, int Height);
void ReSizeGLScene(int Width, int Height);
void DrawGLScene();
void IdleStuff();
void keyPressed(unsigned char key, int x, int y);
void mouseProc(int button, int state, int x, int y);
void mouseMotion(int x, int y);

void free_polyline(polyline_t *p);
void free_block(block_t *b);
void free_entity(entity_t* e);

polyline_t* copy_polyline(polyline_t *p);
block_t* copy_block(block_t *b);

void del_from_list(list_t **list);
void add_to_list(list_t** list, void* newitem);

list_t* read_dxf(char* filename);
int read_colors(char *filename);

#define ESCAPE 27

extern int window;
extern list_t* ents;

#endif
