/** \file generator.c Genrator zada.
 *
 * Program generuje zestawy zada do szeregowania.
 *
 * \author Katarzyna Rafalska
 * \author Jacek Pospychaa
 */

#include<stdlib.h>
#include<stdio.h>
#include<getopt.h>
#include<time.h>

#define BUFLEN 64 /**< Rozmiar bufora. */

/** \brief Gwna funkcja programu.
 *
 * \param argc liczba argumentw, z ktrymi wywoano program
 * \param argv tablica wskanikw do argumentw, z ktrymi wywoano program
 * \return 0 gdy program zakoczy si poprawnie, 1 w przypadku bdu
 */
int main (int argc, char* argv[])
{
  int   lzadan       = 10;  /** Domylnie 10 zada. */
  int   dlugosc_sred = 10;  /** Domylnie dugo 10 */
  int   dlugosc_odch = 3;   /** Odchylenie od redniej dugoci 3 */
  int   termin_min   = 10;  /** Minimalny termin zakoczenia 10 */
  int   termin_max   = 100; /** Maksymalny termin zakoczenia 100 */
  struct timeval s;

  int   i;
  FILE *plik_wy      = NULL;
  char *nazwapliku   = NULL;

  int opt;
  static struct option long_opts[] = {
    {"ltasks", 1, NULL, 'l'},
    {"clen", 1, NULL, 'c'},    
    {"crandomness", 1, NULL, 'r'},
    {"timeout_min", 1, NULL, 't'},
    {"timeout_max", 1, NULL, 'u'},
    {"output", 1, NULL, 'o'},
    {"version", 0, NULL, 'v'},
    {"help", 0, NULL, 'h'}
  };
  
  while ((opt = getopt_long(argc, argv, ":o:l:c:r:t:u:vh", long_opts, NULL)) != -1) 
    {
      switch (opt)
	{
	case ('o'): /* --output PLIK */
	  nazwapliku = (char*) malloc(strlen(optarg)+1);
	  strcpy(nazwapliku, optarg);
	  break;

	case ('l'): /* --lzadan N */
	  lzadan = atoi(optarg);
	  break;

	case ('c'): /* --clen N */
	  dlugosc_sred = atoi(optarg);
	  break;

	case ('r'): /* --crandomness N */
	  dlugosc_odch = atoi(optarg);
	  break;

	case ('t'): /* --timeout_min N */
	  termin_min = atoi(optarg);
	  break;

	case ('u'): /* --timeout_max N */
	  termin_max = atoi(optarg);
	  break;

	case ('v'): /* --version */
	  printf("generator v.2004-10-19\n"
		 "Katarzyna Rafalska (66305), Jacek Pospychaa (66301)\n"
		 );
	  exit(0);
	  break;

	case ('h'): /* --help */
	  printf("generator [opcje]\n"
		 "Program generuje zestawy zada do szeregowania. Opcje:\n"
		 "-l  --ltasks N      Liczba zada\n"
		 "-c  --clen N        Dugo zada\n"
		 "-r  --crandomness N Odchylenie od dugoci zada (+/-)\n"
		 "-t  --timeout_min N Minimalny termin zakoczenia\n"
		 "-u  --timeout_max N Maksymalny termin zakoczenia\n"
		 "-o  --output PLIK   Zapisuje wynik do pliku, zamiast na stdout\n"
		 "-v  --version       Wywietla wersj programu i wychodzi\n"
		 "-h  --help          Pomoc, ktr widzisz\n"
		 );
	  exit(0);
	  break;
	}
    }

  if (nazwapliku != NULL)
    {
      if ((plik_wy = fopen(nazwapliku, "w")) == NULL)
	{
	  perror("Otwarcie pliku wyjciowego");
	  exit(-1);
	}
    }
  else
    {
      plik_wy = stdout;
    }

  fprintf(plik_wy, "%d\n", lzadan);

  gettimeofday(&s, NULL);
  srand(s.tv_usec);

  for (i=0; i<lzadan; i++)
    {
      fprintf(plik_wy, 
	      "%d %d\n",
	      dlugosc_sred - dlugosc_odch + 1 + (int) ( (1+ 2.0*dlugosc_odch*rand()) / (RAND_MAX+1.0) ), 
	      termin_min + (int) ((1.0*(termin_max - termin_min)*rand()) / (RAND_MAX+1.0) ) );
    }

  fclose(plik_wy);
}
