/** \file struktury.h Zbir wsplnych struktur
 *                    
 * Plik zawiera definicje wspdzielonych struktur, oraz deklaracje funkcji 
 * przetwarzania pliku ze zbiorem zada i inne.
 *
 * \author Katarzyna Rafalska
 * \author Jacek Pospychaa
 */

#ifndef STRUKTURY_H

#define STRUKTURY_H 1

/** Struktura reprezentujca zadanie.
 */
struct dane{
  int num;  /**< Numer zadania */
  int c;   /**< Czas trwania zadania */
  int d;   /**< Oczekiwany termin wykonania zadania */
  int l;   /**< Opnienie zadania wzgl. oczekiwanego terminu wykonania*/
  int d_bezc; /**<  kryterium sortowania w quicksorcie */
  int start; /**< kiedy zadanie zostanie rozpoczete */
  struct dane* next; /**< nastepne zadanie na procesorze */
};

struct procesor {
  int wolny_od; /**< od kiedy procesor nic nie robi */
  struct dane* last; /**< wskaznik do ostatniego zadania na procesorze */
};

/** Funkcja pobierajca dane.
 * Funkcja odpowiada za otwarcie pliku z danymi i odczytanie jego zawartoci.
 * Struktura pliku:
 *
 * 
 * LPROC 
 * LTASKS
 * NR C D
 * 0 4564 456
 * 1 55 6776
 * ...
 *
 * LPROC  - liczba procesorw
 * LTASKS - liczba zada
 * NR     - numer zadania
 * C      - czas trwania zadania
 * D      - oczekiwany termin zakoczenia zadania
 *
 * \param nazwa_pliku Nazwa pliku do otwarcia
 * \param lproc wskanik na zmienn, gdzie zostanie zapisana liczba procesorw
 * \param ltasks wskanik na zmienn, gdzie zostanie zapisana liczba procesorw
 * \return Wskanik do obszaru w pamici, gdzie zostay zapisane dane z pliku
 */
struct dane *pobierz_dane(char *nazwa_pliku, int *ltasks)
{
  FILE *plik;
  int licznik=0;
  struct dane *wejscie;
  if ((plik = fopen(nazwa_pliku,"r")) ==NULL)
    {
      perror("Odczyt pliku");
      exit(-1);
    }
  
  
  fscanf(plik,"%d",ltasks);
  wejscie = (struct dane*) malloc((*ltasks) * sizeof(struct dane));
  for (licznik=0; licznik< *ltasks; licznik++)
    {
      wejscie[licznik].num=licznik;
      fscanf(plik,"%d %d",&wejscie[licznik].c,&wejscie[licznik].d);
      wejscie[licznik].d_bezc = wejscie[licznik].d - wejscie[licznik].c;
    }

  return wejscie;
}

/** 
 *
 */
void przydziel_procesor(struct procesor *procesory, struct dane *zadanie)
{
  int min,i;
  min=0;
  
  for (i=1;i<3;i++)
    {
      if (procesory[min].wolny_od>procesory[i].wolny_od) 
	{
	  min=i;
	}
    }

  zadanie->next = NULL;
  zadanie->start = procesory[min].wolny_od;
  procesory[min].wolny_od += zadanie->c;
  zadanie->l = procesory[min].wolny_od - zadanie->d;
  if (procesory[min].last != NULL)
    {
      procesory[min].last->next = zadanie;
    }
  
  procesory[min].last = zadanie;
}

/** Znajduje lmax
 * 
 */
int getlmax(struct dane *wejscie, int ltasks)
{
 int i, lmax;
 for (i=0;i<ltasks;i++) 
 {
   if (i==0 || wejscie[i].l>lmax) {lmax=wejscie[i].l; }
 }
 return lmax;
}

/** Szybkie sortowanie wg. d_bezc
 *
 */
void quicksort_d_bezc(int il, int ih, struct dane *wejscie)
{
  struct dane tmp;
  int med,i,j;
  i=il;
  j=ih;
  med=wejscie[(ih+il)/ 2].d_bezc;
  do {
    while (wejscie[i].d_bezc<med) { i++; }
    while (wejscie[j].d_bezc>med) { j--; }
    if (i<=j)
      {
	memcpy(&tmp, &wejscie[i], sizeof(struct dane));
	memcpy(&wejscie[i], &wejscie[j], sizeof(struct dane));
	memcpy(&wejscie[j], &tmp, sizeof(struct dane));
	i++;
	j--;
      }
  } while (! (i>j));
  if (il<j) { quicksort_d_bezc(il,j,wejscie); } // POPRAWIC
  if (i<ih) { quicksort_d_bezc(i,ih,wejscie); }
}

/** Sortuj wg. numeru czonka struktury okrelonego przez memb
 *
 */
void quicksort_by_memb(int il, int ih, struct dane *wejscie, int memb)
{
  struct dane tmp;
  int med,i,j, moffs;
  i=il;
  j=ih;
  moffs=memb*sizeof(int);
  med=(int)*((int*)((int)(&(wejscie[(ih+il)/ 2])) + moffs));
  do {
    while ((int)(*((int*)((int)(&(wejscie[i])) + moffs))) < med) { i++; }
    while ((int)(*((int*)((int)(&(wejscie[j])) + moffs))) > med) { j--; }
    if (i<=j)
      {
	memcpy(&tmp, &wejscie[i], sizeof(struct dane));
	memcpy(&wejscie[i], &wejscie[j], sizeof(struct dane));
	memcpy(&wejscie[j], &tmp, sizeof(struct dane));
	i++;
	j--;
      }
  } while (! (i>j));
  if (il<j) { quicksort_by_memb(il,j,wejscie, memb); } // POPRAWIC
  if (i<ih) { quicksort_by_memb(i,ih,wejscie, memb); }
}

void stan(struct dane *wejscie, int lproc, int ltasks)
{
  int i=0, p=0;
  struct dane *tmp;
  while (i<ltasks && p<lproc)
  {
    if (wejscie[i].start==0)
    {
      printf("%d (t%d c=%d d=%d l=%d s=%d) ", p, wejscie[i].num, wejscie[i].c, wejscie[i].d, wejscie[i].l, wejscie[i].start);
      tmp = wejscie[i].next;
      while (tmp != NULL)
      {
        printf("(t%d c=%d d=%d l=%d s=%d) ", tmp->num, tmp->c, tmp->d, tmp->l, tmp->start);
        tmp = tmp->next;
      }
      printf("\n");
      p++;
    }
    i++;
  }
  printf("\n");

}

#endif
