/** \file szereguj.c Szeregowanie zada.
 *
 * Program szereguje zadania typu Pn||Lmax, gdzie
 * n to liczba identycznych procesorw.
 * Celem uszeregowania jest minimalizacja opnienia
 * przy wykonywaniu zada.
 *
 * \author Katarzyna Rafalska
 * \author Jacek Pospychaa
 */

#include <stdio.h>
#include "struktury.h"

// wypisywac duzo szczegolow. by nie wypisywal, zakomentowac
#define VERBOSE 1

/** Szybkie sortowanie.
 *
 */
void quicksort(int il, int ih, struct dane *wejscie)
{
  struct dane tmp;
  int med,i,j;
  i=il;
  j=ih;
  med=wejscie[(ih+il)/ 2].d_bezc;
  do {
    while (wejscie[i].d_bezc<med) { i++; }
    while (wejscie[j].d_bezc>med) { j--; }
    if (i<=j)
      {
	memcpy(&tmp, &wejscie[i], sizeof(struct dane));
	memcpy(&wejscie[i], &wejscie[j], sizeof(struct dane));
	memcpy(&wejscie[j], &tmp, sizeof(struct dane));
	i++;
	j--;
      }
  } while (! (i>j));
  if (il<j) { quicksort(il,j,wejscie); } // POPRAWIC
  if (i<ih) { quicksort(i,ih,wejscie); }
}

/** Zamien miejscami wartosci zmiennych a i b
 */
void xchg(int *a, int *b)
{
  int tmp;
  tmp = *a;
  *a = *b;
  *b = tmp;
}

/** Oblicz ponownie opoznienie zadania i opoznienia zadan po nim nastepujacych
 *  tak dlugo, az next bedzie rowne NULL
 */
void update_l(struct dane *task)
{
  task->l = task->start + task->c - task->d;
  if (task->next != NULL)
  {
    task->next->start = task->start + task->c;
    update_l(task->next);
  }
}

/** Zamiana zadan miejscami
 * zamienia miejscami pola c, d w zadaniach, oraz oblicza l tych zadan i aktualizuje l i start 
 * zadan po nich nastepujacych
 */
void exchange(struct dane *task1, struct dane *task2)
{
  xchg(&task1->num, &task2->num);
  xchg(&task1->c, &task2->c);
  xchg(&task1->d, &task2->d);
  update_l(task1);
  update_l(task2);  
}

/** Jesli para (t1,t2) jest w tablicy tabu, zwraca 1,
 *  w przeciwnym wypadku zwraca 0
 */
int in_tabu(int tabu[][2], int tabusize, int tabu_index, int t1, int t2)
{
  int i=0, num;
  
  do
  {
    num = (tabu_index + i) % tabusize;
    i++;
  } while (i<tabusize && (tabu[num][0] != t1 || tabu[num][1] != t2));

  
  if (i==tabusize)
  {
    return 0;
  }
  else
  {
    return 1;
  }
}

/** Dodaje pare (t1,t2) do tabu i zwraca nowa wartosc tabu_index
 */
int add_to_tabu(int tabu[][2], int tabusize, int tabu_index, int t1, int t2)
{
  tabu[tabu_index][0] = t1;
  tabu[tabu_index][1] = t2;
  
  return (tabu_index + 1) % tabusize;
}

/** Algorytm Tabu Search
 * iter - liczba iteracji algorytmu
 * tabusize - wielkosc kolejki tabu
 */
int tabu_search(int ltasks, struct dane *wejscie, int tabusize)
{
  int i, t1,t2, lmax, newlmax, bestnewlmax, bestexchange[2];
  int tabu[tabusize][2];
  int tabu_index = 0;
  
#ifdef VERBOSE
  printf("%d elementowego zbioru zadan, wielkosc listy tabu %d\n", ltasks, tabusize);
#endif
  do
  {
    lmax = getlmax(wejscie, ltasks);
    bestnewlmax = lmax;
#ifdef VERBOSE
    printf("ITERACJA:\nbiezacy lmax: %d\n", lmax);
#endif
          
    for (t1=0; t1<ltasks-1; t1++)
    {
      for (t2=t1+1; t2<ltasks; t2++)
      {
        if (! in_tabu(tabu, tabusize, tabu_index, t1, t2))
	{
          exchange(&wejscie[t1], &wejscie[t2]);
          newlmax = getlmax(wejscie, ltasks);
#ifdef VERBOSE
  printf("\tpo zmianie %d z %d, lmax: %d\n", t1, t2, newlmax);
#endif
	  if (newlmax < bestnewlmax)
	  {
	    bestexchange[0]=t1;
	    bestexchange[1]=t2;
	    bestnewlmax = newlmax;
	  }
	  exchange(&wejscie[t1], &wejscie[t2]);
        }
      }
    }
#ifdef VERBOSE
    printf("najlepszy lmax w zasiegu jednego ruchu: %d\n", bestnewlmax);
#endif
   
    if (bestnewlmax < lmax)
    {
#ifdef VERBOSE
      printf("dokonuje zamiany %d z %d\n", t1, t2);
#endif
      t1 = bestexchange[0];
      t2 = bestexchange[1];      
      exchange(&wejscie[t1], &wejscie[t2]);
      tabu_index = add_to_tabu(tabu, tabusize, tabu_index, t1, t2);
    }
    
  } while (bestnewlmax < lmax);
  
  return lmax;
}

/*
szereguj(lprocesorow, dane)
{
    sortuje dane wg d od najmniejszego do najwiekszego // QUICKSORT
    przydziela do procesorow i oblicza opoznienie jednoczesnie i 
        pamieta najw.opoznienie
}
*/
int szereguj(int ltasks,struct dane *wejscie)
{
 int i,lmax,pr,lproc=3;
 struct procesor *procesory;
 procesory=(struct procesor*)malloc(lproc*sizeof(struct procesor));
 for(i=0;i<lproc;i++)
 {
   procesory[i].wolny_od=0;
   procesory[i].last = NULL;
 }
 
 quicksort(0, ltasks-1, wejscie);
 //quicksort_by_memb(0, ltasks-1, wejscie, 4);
 for (i=0;i<ltasks;i++) 
 {
   przydziel_procesor(procesory, &wejscie[i]);
 }
 
#ifdef VERBOSE
 stan(wejscie, lproc, ltasks);
#endif   

 lmax = tabu_search(ltasks, wejscie, 5);

#ifdef VERBOSE
 stan(wejscie, lproc, ltasks);
#endif 

 free(procesory);
 return lmax;
}

/*
main()
*/
int main(int argc, char *argv[]) {
  struct dane *wejscie;
  int lproc=3,ltasks,i,lmax;
  if (argc<=1)
  {
    printf("Podaj argument\n");
    exit(1);
  }
 wejscie = pobierz_dane(argv[1],&ltasks);
 lmax=szereguj(ltasks,wejscie);

 printf("%d\n",lmax);

 free(wejscie);
 
}

