% 26. combination
combinatio(_, [], []).
combinatio(N, [H|In], [H|Out]) :- combinatio(N, In, Out).

% 25. random permutation of elements from a list
rnd_permu(In, Out) :- length(In, Size),rnd_select(In, Size, Out).

% 24. lotto. draw N different random numbers from a set 1..M
lotto(N,Max,Out) :- range(1,Max,Range),rnd_select(Range,N,Out).

% 23. extract a given number of randomly selected items from a list
rnd_select(_, 0, []).
rnd_select(In, N, [R|Out]) :- length(In,Max1),Max is Max1+1,random(1,Max,Rindex),remove_at(R,In,Rindex,In1),N1 is N-1,rnd_select(In1,N1,Out).

% 22. create a list containing all integers within a range
range(From, To, [From|L]) :- From<To, From1 is From+1, range(From1, To, L).
range(From, From, [From]).

% 21. insert an element at a given position into a list
insert_at(Element, InList, Position, OutList) :- remove_at(Element, OutList, Position, InList).

% 20. remove the K'th element from a list
remove_at(X, [X|T], 1, T1) :- remove_at(X, T, 0, T1).
remove_at(E, [X|T], N, [X|T1]) :- N1 is N-1, remove_at(E, T, N1, T1).
remove_at(_, [], _, []).

% 19. rotate a list N elements to the left
rotate(In, N, Out) :- N<0,!, length(In, X), N1 is X+N, rotate(In, N1, Out).
rotate(In, N, Out) :- length(In,X), X=<N,!, N1 is N mod X, rotate(In, N1, Out).
rotate(In, N, Out) :- split(In,N,In1,In2), append(In2,In1,Out).

% 18. extract a slice between I'th and K'th
slice(_,1,0,[]).
slice([H|T],1,K,[H|T1]) :- K1 is K-1, slice(T,1,K1,T1).
slice([_|T],I,K,T1) :- I1 is I-1, K1 is K-1, slice(T,I1,K1,T1).
slice([],_,_,[]). 

% 17. split list into two parts, length of first part is given
split([X|L],0,L1,[X|L2]) :- split(L,0,L1,L2).
split([X|L],L1size,[X|L1],L2) :- L1size1 is L1size-1, split(L, L1size1, L1, L2).
split([],_,[],[]).

% 16. drop every n'th element of a list
drop(L,N,Z) :- dropc(L,N,N,Z).
dropc([_|T],1,N,Z) :- dropc(T,N,N,Z).
dropc([X|T],K,N,[X|Z]) :- K1 is K-1,dropc(T,K1,N,Z).
dropc([],_,_,[]).

% 15. duplicate elements of a list a given number of times
ktimes(_, 0, []).
ktimes(A, K, [A|T]) :- K1 is K-1, ktimes(A, K1, T).
dupli([X|T], N, T2) :- ktimes(X, N, XList), dupli(T, N, TList), append(XList, TList, T2).
dupli([], _, []).

% 14. duplicate elements of a list
dupli([X|T], [X,X|T2]) :- dupli(T, T2).
dupli([], []).

% 13. encode list, almost like in 10, direct solution
encode_direct([A,A,A|T],[[N,A]|T1]) :- encode_direct([A,A|T], [[N1,A]|T1]),N is N1+1.
encode_direct([A,A,B|T],[[2,A]|T1]) :- encode_direct([B|T], T1).
encode_direct([A,B,B|T],[A|T1]) :- encode_direct([B,B|T], T1).
encode_direct([A,B,C|T], [A|T1]) :- encode_direct([B,C|T], T1).
encode_direct([A,A], [[2,A]]).
encode_direct([A,B], [A,B]).
encode_direct([A], [A]).

% 12. decode run-length encoded list
decode_item([1,A], [A]).
decode_item([N,A], [A|T]) :- N1 is N-1, decode_item([N1,A],T).
decode([], []).
decode([H|T], Out) :- list(H),!,decode_item(H, List),decode(T,Out1),append(List,Out1,Out).
decode([H|T], Out) :- decode(T,Out1),append([H],Out1,Out).

% 11. do things like in problem 10, but treat differently elements without duplicates
mpack_count([],[]).
mpack_count([[L]|T], [L|B]) :- mpack_count(T,B),!.
mpack_count([L|T], [[Ecount,Element]|B]) :- length(L, Ecount),[Element|_]=L,mpack_count(T,B).
encode_modified(A,B) :- pack(A,B1), mpack_count(B1,B).

% 10. encode list, indirect solution - by using previous exercise
pack_count([],[]).
pack_count([L|T], [[Ecount,Element]|B]) :- length(L, Ecount),[Element|_]=L,pack_count(T,B).
encode(A,B) :- pack(A,B1), pack_count(B1,B).

% 9. Pack consecutive duplicates of list elements into sublists
pack([X], [[X]]).
pack([H,H|T], [H1|T1]) :- pack([H|T], [H2|T1]),append([H],H2,H1),!.
pack([X,H|T], [[X],H1|T1]) :- pack([H|T], [H1|T1]).

% 8. Eliminate consecutive duplicates of list elements
compress([X], [X]).
compress([H,H|T], T1) :- compress([H|T], T1), !.
compress([H,Y|T], [H|T1]) :- compress([Y|T], T1).

% 7. Flatten a list
flatten([], []).
flatten([H|T], Flat) :- list(H), !, flatten(H, Tail1), flatten(T, Tail2), append(Tail1, Tail2, Flat).
flatten([H|T], [H|Flat]) :- flatten(T, Flat).

% 6. Palindrome
palindrome(L) :- reverse(L, LR), L=LR.

% 5. Reverse a list
my_reverse([H|T1], L) :- reverse(T1, L1), append(L1,[H],L).

% 4. Number of elements of a list
listsize(0, []).
listsize(N, [_|T]) :- listsize(N1,T), N is N1+1.

% 3. K'th element
element_at(X, [X|_], 1).
element_at(X, [_|T], N) :- N1 is N-1, element_at(X,T,N1).

% 2. last but one
last_but_one(X, [X,_]).
last_but_one(X, [_|T]) :- last_but_one(X,T).

% 1. last element
my_last(X, [X]).
my_last(X, [_|T]) :- my_last(X,T).
