/****
   LTPS - Linux Text Presentation System
                                       ****/

#ifndef TMP_TAG
#define TMP_TAG

#include <vector>

#include "style.h"

// tag types (only possible values fot tagtype_t)
#define TAG_SLIDE    0
#define TAG_UNKNOWN  1
// text structure
#define TAG_P        2
#define TAG_SPAN     3
#define TAG_BR       4
#define TAG_HR       5
#define TAG_UL       6
#define TAG_OL       7
#define TAG_LI       8
// non text, block objects
#define TAG_IMG      9
// tables
#define TAG_TABLE    10
#define TAG_TR       11
#define TAG_TD       12
// text appearance
#define TAG_B        13
#define TAG_I        14
#define TAG_U        15
#define TAG_H1       16
#define TAG_H2       17
#define TAG_H3       18

using namespace std;

typedef int tagtype_t;

class Tag {
 protected:
  int type;           // tag type (one of TAG_* defs)
  int layer;          // layer numer, 0 is default, first, higher are later, lower are former
  Style style;        // style definition
  char* content;      // regular text kept in this tag
  Tag* parent;        // pointeg to parent tag
  vector <Tag*> tags; // child tags

 public:

  Tag();              // most primitive constructor setting everything to null
  Tag(Tag* parent);   // this constructor sets the parent tag of currently created one
  ~Tag();

  Tag* getparent(void);  // returns pointer to parent tag

  tagtype_t settype(char* newtype);     // sets type basing. converts string to tagtype_t and calls settype(tagtype_t), 
                                        // returns -1 if wrong type
  tagtype_t settype(tagtype_t newtype); // sets type to newtype <- validates if tag with this type can be used with current parent, 
                                        // returns -1 if wrong type
  tagtype_t gettype(void);              // returns private field type 

  void setattrs(const char** attrs);
  
  int getlayer(void);
  int getmaxlayer(void);

  void setcontent(char* newcontent, int len);    // sets content to newcontent, forgetting old content (if any)
                                                 // if tags is not empty, new content is added as TAG_SPAN tag to keep the order
  void appendcontent(char* newcontent, int len); // appends newcontent to current content. if tags is not empty, 
                                                 // new content is added as TAG_SPAN tag to keep the order
  char* getcontent(void);   // returns raw content, or NULL if there are tags inside
  char getfirstchar(void);  // returns first char of content, with recursive subtags check
  char getlastchar(void);  // returns last char of content, with recursive subtags check

  void addtag(Tag* tag); // adds a child tag. if current Tag already has some content it is changed to TAG_SPAN and added
                         // first. later if any content is added, it's added as other TAG_SPAN tags to keep the order
                         // of parts being tags and raw content.

  int minwidth(void);        // returns minimal width required to display formatted
  //  int minheight(int width);  // returns minimal height required to display formatted with specified width
  int length(void);          // returns lenght of Tag content and child tags

  int isblock(void);     // tells wether tag is a block and requires whole width. If it is not block, it can be put floating in text

  char*** Tag::getformatted(char*** current, int width, int height, int wx, int wy, int wwidth, int wheight, int *xoffset, int *yoffset, int layer);
  char*** getformatted(int width, int height, int layer); // returns 2d array [width*height] with formatted content from specified layer
};

#endif
