/****
   LTPS - Linux Text Presentation System
                                       ****/

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <ncurses.h>
#include <time.h>

#include "transition.h"

using namespace std;

Transition::Transition()
{
}

Transition::Transition(char* opts)
{
}

int Transition::execute(char*** data, int width, int height)
{
  int y;
  char** text = NULL, **effects = NULL;

  text = data[0];
  effects = data[1];
  
  for (y=0; y<height; y++)
    mvprintw(y, 0, "%s", text[y]);

  refresh();

  return 0;
}

int Trans_Up2Down::execute(char*** data, int width, int height)
{
  int y=0;
  struct timespec sleeptime;
  char** text = NULL, **effects = NULL;

  text = data[0];
  effects = data[1];
  
  for (y=0; y<height; y++)
    {
      sleeptime.tv_sec = 0;
      sleeptime.tv_nsec = 10000000;

      nanosleep(&sleeptime, NULL);
      mvprintw(y, 0, "%s", text[y]);
      refresh();
    }

  return 0;
}

int Trans_Left2Right::execute(char*** data, int width, int height)
{
  int x=0, y=0;
  struct timespec sleeptime;
  char** text = NULL, **effects = NULL;

  text = data[0];
  effects = data[1];

  for (x=0; x<width; x++)
    {
      sleeptime.tv_sec = 0;
      sleeptime.tv_nsec = 10000000;

      nanosleep(&sleeptime, NULL);
      
      for (y=0; y<height; y++)
	{
	  mvaddch(y, x, text[y][x]);
	}
      refresh();
    }

  return 0;
}

int Trans_Center2Borders::execute(char*** data, int width, int height)
{
  int x0, y0, r, rmax, d;
  struct timespec sleeptime;
  char** text = NULL, **effects = NULL;

  text = data[0];
  effects = data[1];
  
  x0 = width/2;
  y0 = height/2;
  rmax = x0>y0 ? x0 : y0;
  
  for (r=0; r<rmax+1; r++)
    {
      for (d=y0-r; d<y0+r; d++)
	{
	  if ((0 <= d) && (d < height) && (x0+r < width))
	      mvaddch(d, x0+r, text[d][x0+r]);
	  if ((0 <= d) && (d < height) && (0 <= x0-r))
	      mvaddch(d, x0-r, text[d][x0-r]);

	}
      for (d=x0-r; d<=x0+r; d++)
	{
	  if ((0 <= d) && (d < width) && (y0+r < height))
	      mvaddch(y0+r, d, text[y0+r][d]);
	  if ((0 <= d) && (d < width) && (0 <= y0-r))
	      mvaddch(y0-r, d, text[y0-r][d]);
	}

      refresh();

      sleeptime.tv_sec = 0;
      sleeptime.tv_nsec = 10000000;
      
      nanosleep(&sleeptime, NULL);
    }

  return 0;
}

int Trans_MoveFromLeft::execute(char*** data, int width, int height)
{
  int x, movx, y;
  struct timespec sleeptime;
  char** text = NULL, **effects = NULL;

  text = data[0];
  effects = data[1];
  
  for (movx=1; movx<=width; movx++)
    {      
      for (x=0; x<movx; x++)
	{
	  for (y=0; y<height; y++)
	    {
	      mvaddch(y, x, text[y][width-movx+x]);
	    }
	}
      refresh();

      sleeptime.tv_sec = 0;
      sleeptime.tv_nsec = 10000000;
      nanosleep(&sleeptime, NULL);
    }

  return 0;
}

char Trans_Matrix::randchar()
{
  int i = (int)(rand()*35.0/(RAND_MAX + 1.0));
  return (i<25) ? i+97 : i-25+48;
}

int Trans_Matrix::execute(char*** data, int width, int height)
{
  struct raindrop {int x0, yup, ydown, length, speed, tick;} *raindrops;

  int i, x=0, y=0, drops=width, dropsleft=0;
  //  struct timespec sleeptime;
  int left=0;
  char** text = NULL, **effects = NULL;
  char** ready;

  start_color();
  init_pair(4, COLOR_GREEN, COLOR_BLACK);
  attron(COLOR_PAIR(4));

  text = data[0];
  effects = data[1];
  
  srand(time(NULL));

  raindrops = (struct raindrop*) malloc(drops*sizeof(struct raindrop));

  ready = (char**) malloc(height* sizeof(char*));
  for (i=0; i<height; i++)
    {
      ready[i] = (char*) malloc(width);
      bzero(ready[i], width);
    }

  for (x=0; x<width; x++)
    for (y=0; y<height; y++)
      left += (text[y][x]==' ') ? 0 : 1;

  for (i=0; i<drops; i++)
    raindrops[i].yup=height+1;

  while (left>0 || dropsleft>0)
    {
      for (i=0; i<drops; i++)
	{
	  if (raindrops[i].yup>height) // new raindrop
	    {
	      if (left==0)
		{
		  dropsleft--;
		  raindrops[i].yup = raindrops[i].tick = raindrops[i].length = raindrops[i].ydown = 0;
		}
	      else
		{
		  dropsleft++;
		  raindrops[i].length = (int)(1.0*rand()*height/(RAND_MAX +1.0));
		  raindrops[i].x0 = (int)(1.0*width*rand()/(RAND_MAX +1.0));
		  raindrops[i].speed = 1 + (int)((10.0*rand())/(RAND_MAX +1.0));
		  raindrops[i].ydown = -(int)(1.0*height*rand()/(RAND_MAX +1.0));
		  raindrops[i].yup = raindrops[i].tick = 0;
		}
	    }
	  else
	    {
	      raindrops[i].tick = (raindrops[i].tick + 1) % raindrops[i].speed;
	      if (raindrops[i].tick == 0)
		{
		  raindrops[i].ydown++;

		  if (raindrops[i].ydown>raindrops[i].length)
		    {
		      for (y=raindrops[i].yup-raindrops[i].length; y<raindrops[i].yup; y++)
			{
			  if (0<=y && y<height && 0 <= raindrops[i].x0 && raindrops[i].x0<width)
			    if ((! ready[y][raindrops[i].x0]) || text[y][raindrops[i].x0]==' ')
			      mvaddch(y, raindrops[i].x0, ' ');
			}
		      raindrops[i].yup++;
		    }
		  
		  if (raindrops[i].yup<raindrops[i].ydown)
		    {
		      for (y=raindrops[i].yup; y<raindrops[i].ydown; y++)
			{			    
			  if (0<=y && y<height && 0 <= raindrops[i].x0 && raindrops[i].x0<width)
			    {
			      if (! ready[y][raindrops[i].x0])
				{
				  mvaddch(y, raindrops[i].x0, randchar());
				  if (text[y][raindrops[i].x0]!=' ' && rand()>0.95)
				    {
				      ready[y][raindrops[i].x0]=1;
				      left--;
				    }
				}
				  else
				    mvaddch(y, raindrops[i].x0, text[y][raindrops[i].x0]);
			    }
			}
		    }
		}
	  
	      refresh();  
	    }
	}
    }

  attroff(COLOR_PAIR(4));

  return 0;
}
