/*
  obsluga sieci wymiany plikow wg. protokolu grupy I4, z srody 8:00
  Jacek Pospychaa (66301)
 */

#include <stdlib.h>
#include <signal.h>
#include <sys/ipc.h>
#include <sys/msg.h>

#include <regex.h>

#include "66301_common.h"
#include "66301_EXTENDEDprotocol.h"

/**************************************
 * bezpieczestwo, rne
 **************************************/

/* zwraca dugo null-terminated stringa, ale nie wiksz ni n.
 */
size_t strnlen(char* s, int n)
{
  DEBUGMSG(TRACE_DBG, ("strnlen\n"));
  if (memchr(s, '\0', n) == 0)
    {
      return n;
    }
  else
    {
      return ((int)memchr(s, '\0', n) - (int)s);
    }
}

/* wrapper na malloc, koczcy dziaanie programu, jeli zabraknie
   pamici
 */

void *xmalloc(size_t size)
{
  DEBUGMSG(TRACE_DBG, ("xmalloc\n"));
  void* mem;
  DEBUGMSG(X_DBG, ("xmalloc(%d): ", size));
  ASSERT(size>0,);
  if ((mem = (void*) malloc(size)) == NULL)
    {
      perror("malloc");
      CALL_EXIT;
    }
  DEBUGMSG(X_DBG, ("allocated at %x\n", mem));
  return mem;
}

void xfree(void* ptr)
{
  DEBUGMSG(TRACE_DBG, ("xfree\n"));
  DEBUGMSG(X_DBG, ("free %d\n", ptr));
  free(ptr);
}

int ereg(char* pattern, char* s)
{
  regex_t* preg;
  int retval;

  regcomp(preg, pattern, REG_EXTENDED|REG_NOSUB);
  retval = regexec(preg, s, 0, NULL, 0);
  regfree(preg);

  return retval;
}

/**************************************
 * tworzenie i usuwanie kolejek
 **************************************/

/* Tworzy kolejk komunikatw i zwraca jej msgid. Jeli taka kolejka 
   ju istnieje, lub wystpi inny bd uniemoliwiajcy jej stworzenie,
   program koczy dziaanie.
 */
int create_queue(int qid)
{
  DEBUGMSG(TRACE_DBG, ("create_queue\n"));
  DEBUGMSG(MESSAGE_DBG, ("creating exclusive queue\n"));
  int msgid;

  if ((msgid = msgget(qid, 0664|IPC_CREAT|IPC_EXCL)) == -1)
    {
      perror("create_queue msgget");
      CALL_EXIT;
    }
  return msgid;
}

/* Usuwa kolejk, jeli istnieje. Jeli wystpi bd, zostanie on wywietlony,
   ale nie powoduje zakoczenia programu.
 */
void close_queue(int msgid)
{
  DEBUGMSG(TRACE_DBG, ("close_queue\n"));
  DEBUGMSG(MESSAGE_DBG, ("closing queue\n"));

  if (msgid != -1)
    {
      if (msgctl(msgid, IPC_RMID, NULL) == -1)
	{
	  perror("close_queue msgctl IPC_RMID");
	}
    }
}

/**************************************
 * komunikacja
 **************************************/

/* wrapper na msgrcv obsugujcy moliwo niepowodzenia podczas
   odbierania wiadomoci (powd: usunicie kolejki/sygna)
   Zwraca zawsze ssize_t z rozmiarem odebranych danych
*/
ssize_t xmsgrcv(int msqid, void* msgp, size_t msgsz, long msgtyp, int msgflg)
{
  DEBUGMSG(TRACE_DBG, ("xmsgrcv\n"));
  ssize_t retval;

  if ((retval = msgrcv(msqid, msgp, msgsz, msgtyp, msgflg)) == -1)
    {
      perror("xmsgrcv");
      CALL_EXIT;
    }

#if (DEBUG&MESSAGE_DBG)
  struct message *msg = (struct message*)msgp;
  char arg1[PROTO_ARGSIZE+1] = {0};
  char arg2[PROTO_ARGSIZE+1] = {0};
  char body[PROTO_MSGBODYSIZE+1] = {0};
  strncpy(arg1, msg->argument1, PROTO_ARGSIZE);
  strncpy(arg2, msg->argument2, PROTO_ARGSIZE);
  strncpy(body, msg->messagebody, PROTO_MSGBODYSIZE);
  printf("msgrcv (mtype %d, size: %d, msgtyp: %ld)\n"
	 "\tpart:     %d/%d\n"
	 "\tclientid: %d\n"
	 "\tcmd:      %d\n"
	 "\targ1:     %s\n"
	 "\targ2:     %s\n"
	 "\tbody len: %d\n", msqid, msgsz, msgtyp, msg->partnumber, msg->partscount, msg->clientid, msg->command, arg1, arg2, strnlen(body, PROTO_MSGBODYSIZE));
#endif

  return retval;
}

/* wrapper na msgsnd obsugujcy moliwo niepowodzenia podczas
   wysyania wiadomoci (powd: usunicie kolejki/sygna)
   Zwraca zawsze 0
*/
int xmsgsnd(int msqid, void* msgp, size_t msgsz, int msgflg)
{
  DEBUGMSG(TRACE_DBG, ("xmsgsnd\n"));

#if (DEBUG&MESSAGE_DBG)
  struct message *msg = (struct message*)msgp;
  char arg1[PROTO_ARGSIZE+1] = {0};
  char arg2[PROTO_ARGSIZE+1] = {0};
  char body[PROTO_MSGBODYSIZE+1] = {0};
  strncpy(arg1, msg->argument1, PROTO_ARGSIZE);
  strncpy(arg2, msg->argument2, PROTO_ARGSIZE);
  strncpy(body, msg->messagebody, PROTO_MSGBODYSIZE);
  printf("msgsnd (mtype %d, size: %d)\n"
	 "\tpart:     %d/%d\n"
	 "\tclientid: %d\n"
	 "\tcmd:      %d\n"
	 "\targ1:     %s\n"
	 "\targ2:     %s\n"
	 "\tbody len: %d\n", msqid, msgsz, msg->partnumber, msg->partscount, msg->clientid, msg->command, arg1, arg2, strnlen(body, PROTO_MSGBODYSIZE));
#endif

  if (msgsnd(msqid, msgp, msgsz, msgflg) == -1)
    {
      perror("xmsgsnd");
      CALL_EXIT;
    }
  return 0;
}

/**************************************
 * protok
 **************************************/

/* zwraca ile argumentw moe mie komenda (0 - brak, 1, lub 2)
 */
int command_has_args(int command)
{
  DEBUGMSG(TRACE_DBG, ("command_has_args\n"));
  command = command % 10; // dla bdu liczba argumentw jest taka sama 
  char results[9] = {2, 0, 2, 2, 1, 0, 0, 1, 2};
  return results[command-1];
}

/* zwraca 1, jeli komenda moe zawiera tre, lub 0 jeli nie
 */
int command_has_body(int command)
{
  DEBUGMSG(TRACE_DBG, ("command_has_body\n"));
  char results[9] = {0, 0, 1, 1, 1, 1, 1, 0, 0};
  if (command>10) // wszystkie errory maj tre
    {
      return 1;
    }
  return results[command-1];
}

/* wysyanie wiadomoci. wszystkie pola musz by wypenione.
   funkcja zawsze zwraca 1.
*/
int send_message(int msgid, complete_msg* message)
{
  DEBUGMSG(TRACE_DBG|MESSAGE_DBG, ("send_message\n"));

  struct message msg;
  int bytes_sent = 0;
  ASSERT(message!=NULL,);
  
  bzero(&msg, PROTO_MSGSIZE);
  msg.clientid = message->clientid;
  msg.command = message->command;

  if ((command_has_args(msg.command) >= 1) && (message->argument1!=NULL))
    {
      strncpy(msg.argument1, message->argument1, PROTO_ARGSIZE);
    }
  if ((command_has_args(msg.command) >= 2) && (message->argument2!=NULL))
    {
      strncpy(msg.argument2, message->argument2, PROTO_ARGSIZE);
    }
  
  if (command_has_body(msg.command) && (message->messagebody!=NULL) && (message->messagebody[0]!='\0'))
    {
      msg.partscount = (strlen(message->messagebody) / PROTO_MSGBODYSIZE) + 
	((strlen(message->messagebody) % PROTO_MSGBODYSIZE) > 0);

      while (bytes_sent < strlen(message->messagebody))
	{
	  strncpy(msg.messagebody, &(message->messagebody[bytes_sent]), PROTO_MSGBODYSIZE);
	  bytes_sent += strnlen(msg.messagebody, PROTO_MSGBODYSIZE);
	  xmsgsnd(msgid, &msg, PROTO_MSGSIZE-sizeof(long), 0);
	  msg.partnumber++;
	}
    }
  else
    {
      msg.partscount = 1;
      xmsgsnd(msgid, &msg, PROTO_MSGSIZE-sizeof(long), 0);
    }

  return 1;
}

/* Pobiera wiadomo od uytkownika o okrelonym clientid, jeli command!=0, wwczas
   bierze tylko wiadomo z okrelon komend.
   Struktura message musi ju istnie, ale pola w niej zostan wypenione (i zaalokowane)
   przez funkcj.
   Wiadomo po odczytaniu mona usun przy pomocy free_message();
 */
int receive_message(int msgid, long clientid, int command, complete_msg* message, int flag)
{
  DEBUGMSG(MESSAGE_DBG|TRACE_DBG, ("receive_message\n"));
  struct message msg_1st;
  struct message msg;
  int part_to_receive;
  ASSERT(message!=NULL,);

  // upewnijmy si, e wzita zostanie pierwsza cz wiadomoci, oczekiwnego typu, 
  // a wszystkie pozostae wrc spowrotem do kolejki
  while (xmsgrcv(msgid, &msg_1st, PROTO_MSGSIZE-sizeof(long), clientid, 0)
	 && ((msg_1st.partnumber != 0) || (command!=0 && msg_1st.command!=command)))
    {
      if (flag&KEEP_UNWANTED)
	xmsgsnd(msgid, &msg_1st, PROTO_MSGSIZE-sizeof(long), 0);
    }

  // kopiujemy pole clientid, command i argumenty
  message->clientid = msg_1st.clientid;
  message->command = msg_1st.command;
  if ((command_has_args(msg_1st.command) >= 1) && (msg_1st.argument1[0] != '\0'))
    {
      message->argument1 = (char*) xmalloc(strnlen(msg_1st.argument1, PROTO_ARGSIZE)+1); 
      // +1 bo trzeba miejsca na dugo stringa i '\0' wic nawet, jakby w argument1 nie byo '\0' to i tak zostanie dodane
      bzero(message->argument1, strnlen(msg_1st.argument1, PROTO_ARGSIZE)+1);
      strncpy(message->argument1, msg_1st.argument1, strnlen(msg_1st.argument1, PROTO_ARGSIZE));
    }
  else
    {
      message->argument1 = NULL;
    }

  if ((command_has_args(msg_1st.command) >= 2) && (msg_1st.argument2[0] != '\0'))
    {
      message->argument2 = (char*) xmalloc(strnlen(msg_1st.argument2, PROTO_ARGSIZE)+1); 
      // +1 bo trzeba miejsca na dugo stringa i '\0' wic nawet, jakby w argument1 nie byo '\0' to i tak zostanie dodane
      bzero(message->argument2, strnlen(msg_1st.argument2, PROTO_ARGSIZE)+1);
      strncpy(message->argument2, msg_1st.argument2, strnlen(msg_1st.argument2, PROTO_ARGSIZE));
    }
  else
    {
      message->argument2 = NULL;
    }

  // kopiujemy tre
  if (command_has_body(msg_1st.command) && (msg_1st.messagebody[0] != '\0'))
    {
      if (strnlen(msg_1st.messagebody, PROTO_MSGBODYSIZE) < PROTO_MSGBODYSIZE)
	{ // jesli wszystko miesci sie w jednym pakiecie
	  message->messagebody = (char*) xmalloc(strnlen(msg_1st.messagebody, PROTO_MSGBODYSIZE)+1); // +1 na '\0' j.w.
	  bzero(message->messagebody, strnlen(msg_1st.messagebody, PROTO_MSGBODYSIZE)+1);
	}
      else { // jesli bedzie wiecej pakietow
	message->messagebody = (char*) xmalloc((msg_1st.partscount * PROTO_MSGBODYSIZE) + 1); // +1 na '\0' j.w.
	bzero(message->messagebody, (msg_1st.partscount * PROTO_MSGBODYSIZE) + 1);
      }
      strncpy(message->messagebody, msg_1st.messagebody, strnlen(msg_1st.messagebody, PROTO_MSGBODYSIZE));
    }
  else
    {
      message->messagebody = NULL;
    }

  part_to_receive = 1;

  // teraz pobieramy czci o numerach 'part_to_receive' we waciwej kolejnoci, tyle
  // ile ich jest
  while (part_to_receive < msg_1st.partscount)
    {
      while (xmsgrcv(msgid, &msg, PROTO_MSGSIZE-sizeof(long), clientid, 0)
	     && (msg.partnumber != part_to_receive ||
		 msg.command != message->command ||
		 (message->argument1!=NULL && strncmp(msg.argument1, message->argument1, PROTO_ARGSIZE) != 0) ||
		 (message->argument2!=NULL && strncmp(msg.argument2, message->argument2, PROTO_ARGSIZE) != 0)
		 )
	     )
	{
	  if (command!=0 && msg_1st.command!=command && (flag&KEEP_UNWANTED))
	    {
	      xmsgsnd(msgid, &msg, PROTO_MSGSIZE-sizeof(long), 0);
	    }
	}

      strncpy((char*)((int)message->messagebody + (int)strlen(message->messagebody)), 
	      msg.messagebody, 
	      strnlen(msg.messagebody, PROTO_MSGBODYSIZE));

      part_to_receive++;
    }

  return 1;
}

/* Zwalnia pola struktury i wstawia NULL. 
   
   Nie wywietla komunikatw, ani nie powoduje wyjcia z programu w adnej sytuacji. 
   Jeli komunikat podany w argumencie nie by zaalokowany, funkcja nic nie robi.
*/
void free_message(complete_msg* message)
{
  DEBUGMSG(TRACE_DBG, ("free_message\n"));
  if (message!=NULL)
    {
      xfree(message->argument1);
      xfree(message->argument2);
      xfree(message->messagebody);
      bzero(message, sizeof(complete_msg));
    }
}

void answer_error(complete_msg* msg, char* error_msg)
{
  DEBUGMSG(TRACE_DBG, ("answer_error\n"));
  msg->command+= 10;
  xfree(msg->messagebody);
  msg->messagebody = error_msg;
  send_message(msg->clientid, msg);
  msg->messagebody = NULL;
#ifdef H66301_SERVER
  log_message(error_msg);
#endif
}

void answer_success(complete_msg* msg)
{
  DEBUGMSG(TRACE_DBG, ("answer_success\n"));
  xfree(msg->messagebody);
  msg->messagebody = NULL;
  send_message(msg->clientid, msg);
}
