/****
   LTPS - Linux Text Presentation System
                                       ****/

#include <stdio.h>
#include <string.h>
#include <iostream>

#include "tag.h"
#include "style.h"

#include "common.h"

using namespace std;

int blank(char c)
{
  return ((c=='\n') || (c==' ') || (c=='\t') || (c=='\r'));
}

char* strip_blanks(char* s, int len)
{
  int si, bufi=0;
  char *buf = new char[len+1];

  for (si=0; si<len; si++)
    {
      if (blank(s[si]))
	{
	  if (bufi>0)
	    {
	      buf[bufi]=' ';
	      bufi++;
	    }
	  
	  while (si+1<len && blank(s[si+1]))
	    si++;

	}
      else
	{
	  buf[bufi] = s[si];
	  bufi++;
	}
    }
  buf[bufi]=0;

  return buf;
}

/* function returns last position of last char c in str, before n,
   or -1 if not found */
int strnrchr(char* s, int n, char c)
{
  if (s==NULL) return -1;
  if ((signed int)strlen(s)<n)
    n=strlen(s)-1;

  while (n>=0 && s[n]!=c)
    n--;

  return n;
}

Tag::Tag()
{
  parent = NULL;
  content = NULL;
  this->style = Style();
}

Tag::Tag(Tag* parent)
{
  //Tag();
  this->parent = NULL;
  content = NULL;
  this->parent = parent;
  this->style = Style();
}

Tag::~Tag()
{
}

Tag* Tag::getparent(void)
{
  return parent;
}

tagtype_t Tag::settype(char* newtype)
{
  tagtype_t type;

  if (strncmp(newtype, "p", 2) == 0)
    type = TAG_P;
  else if (strncmp(newtype, "table", 6) == 0)
    type = TAG_TABLE;
  else if (strncmp(newtype, "tr", 3) == 0)
    type = TAG_TR;
  else if (strncmp(newtype, "td", 3) == 0)
    type = TAG_TD;
  else if (strncmp(newtype, "b", 2) == 0)
    type = TAG_B;
  else if (strncmp(newtype, "i", 2) == 0)
    type = TAG_I;
  else if (strncmp(newtype, "u", 2) == 0)
    type = TAG_U;
  else if (strncmp(newtype, "br", 3) == 0)
    type = TAG_BR;
  else if (strncmp(newtype, "hr", 3) == 0)
    type = TAG_HR;
  else if (strncmp(newtype, "ul", 3) == 0)
    type = TAG_UL;
  else if (strncmp(newtype, "ol", 3) == 0)
    type = TAG_OL;
  else if (strncmp(newtype, "li", 3) == 0)
    type = TAG_LI;
  else if (strncmp(newtype, "h1", 3) == 0)
    type = TAG_H1;
  else if (strncmp(newtype, "h2", 3) == 0)
    type = TAG_H2;
  else if (strncmp(newtype, "h3", 3) == 0)
    type = TAG_H3;
  else if (strncmp(newtype, "img", 4) == 0)
    type = TAG_IMG;
  else if (strncmp(newtype, "span", 4) == 0)
    type = TAG_SPAN;
  else type = TAG_UNKNOWN;

  return settype(type);
}

tagtype_t Tag::settype(tagtype_t newtype)
{
  int valid=0;
  ASSERT(this->parent!=NULL,"Tag with type and without parent? Cant validate type");
  tagtype_t ptype = this->parent->gettype();

  switch (newtype)
    {
    case TAG_TR:
	valid = (ptype==TAG_TABLE);
	break;
    case TAG_TD:
	valid = (ptype==TAG_TR);
	break;
    case TAG_TABLE:
    case TAG_P:
    case TAG_SPAN:
    case TAG_B:
    case TAG_I:
    case TAG_U:
    case TAG_BR:      
    case TAG_HR:      
    case TAG_H1:
    case TAG_H2:
    case TAG_H3:
    case TAG_OL:
    case TAG_UL:
    case TAG_IMG:
      valid = (ptype!=TAG_TABLE && ptype!=TAG_TR && ptype!=TAG_BR);
      break;
    case TAG_LI:
      valid = (ptype == TAG_UL || ptype == TAG_OL);
      break;
    case TAG_UNKNOWN: // can be used anywhere
    default:
      valid=1;
    }

  if (valid)
    {
      switch (newtype) // apply default style
	{
	case TAG_P:
	  style.setmargin(1);
	  style.setmarginbottom(2);
	  break;
	case TAG_TD:
	  style.setmargin(1);
	  break;
	case TAG_HR:
	  style.setmargintop(1);
	  style.setmarginbottom(1);
	  break;
	case TAG_H1:
	  style.setmargin(2);
	  style.setalign(ha_center);
	  // bold
	  break;
	case TAG_H2:
	  style.setmargintop(1);
	  style.setalign(ha_center);
	  break;
	case TAG_H3:
	  style.setmargin(1);
	  style.setmarginleft(3);
	  break;
	case TAG_OL:
	case TAG_UL:
	  style.setmarginleft(1);
	  break;
	case TAG_LI:
	  style.setmarginleft(1);
	  break;
	case TAG_IMG:
	case TAG_TR:
	case TAG_TABLE:
	case TAG_SPAN:
	case TAG_B:
	case TAG_I:
	case TAG_U:
	case TAG_BR:
	break;
	}
    }

  return (valid) ? this->type=newtype : -1;
}

tagtype_t Tag::gettype(void)
{
  return type;
}

void Tag::setcontent(char* newcontent, int len)
{
  if (content != NULL)
    delete content;

  if (tags.size() > 0) // if there are some tags, add content as a tag
    {
      Tag* tmpnewtag = new Tag(this);
      tmpnewtag->settype(TAG_SPAN);
      tmpnewtag->appendcontent(newcontent, len);
      this->addtag(tmpnewtag);
    }
  else
    {
      content = new char[len + 1];
      content = strip_blanks(newcontent, len);
      //      bzero(content, len+1);
      //      strncpy(content, newcontent, len);
    }
}

void Tag::appendcontent(char* newcontent, int len)
{
  if (tags.size() > 0) // if there are some tags, add content as a tag
    {
      Tag* tmpnewtag = new Tag(this);
      tmpnewtag->settype(TAG_SPAN);
      tmpnewtag->appendcontent(newcontent, len);
      this->addtag(tmpnewtag);
    }
  else if (content == NULL)
    {
      setcontent(newcontent, len);
    }
  else
    {
      char* buf = new char[strlen(content) + len + 2];

      bzero(buf, strlen(content) + len + 2);
      snprintf(buf, strlen(content) + len+1, "%s%s", content, newcontent);
      delete content;
      content = strip_blanks(buf, strlen(buf));
      //      content = buf;
    }
}

char* Tag::getcontent(void)
{
  return content;
}

char Tag::getfirstchar(void)
{
  if (content!=NULL)
    return content[0];
  else if (tags.size()>0)
    return tags[0]->getfirstchar();
  else return 0;
}

char Tag::getlastchar(void)
{
  if (content!=NULL)
    return content[strlen(content)-1];
  else if (tags.size()>0)
    return tags[tags.size()-1]->getlastchar();
  else return 0;
}

void Tag::addtag(Tag* tag)
{
  if (content != NULL)
    {
      Tag* tmpnewtag = new Tag(this);
      tmpnewtag->settype(TAG_SPAN);
      tmpnewtag->appendcontent(this->content, strlen(this->content));
      free(content);
      content = NULL;
      tags.push_back(tmpnewtag);
    }

  tags.push_back(tag);
}

int Tag::minwidth(void)
{
  int minwidth;

  switch (type)
    {
    case TAG_TABLE: // minwidth is the maximum of children minwidths
    case TAG_UL:
    case TAG_OL:
      {
	unsigned int i;
	minwidth = style.getmarginleft() + style.getmarginright();;
	for (i = 0; i<this->tags.size(); i++)
	  {
	    if (this->tags[i]->minwidth() + style.getmarginleft() + style.getmarginright() > minwidth)
	      minwidth = this->tags[i]->minwidth() + style.getmarginleft() + style.getmarginright();;
	  }
	break;
      }
    case TAG_TR: // minwidth is the sum of columns minwidths (children)
      {
	unsigned int i;
	minwidth = style.getmarginleft() + style.getmarginright();
	for (i = 0; i<this->tags.size(); i++)
	  minwidth += this->tags[i]->minwidth();
	break;
      }
    case TAG_TD: // default minwidth for floating elements
    case TAG_P:
    case TAG_SPAN:
    case TAG_B:
    case TAG_I:
    case TAG_U:
    case TAG_BR:
    case TAG_HR:
    case TAG_LI:
    case TAG_H1:
    case TAG_H2:
    case TAG_H3:
      minwidth = 3 + style.getmarginleft() + style.getmarginright();
      break;
    case TAG_IMG:
      minwidth = 5 + style.getmarginleft() + style.getmarginright();
      break;
    case TAG_UNKNOWN:
    default:      
      minwidth = style.getmarginleft() + style.getmarginright();;
    }

  return minwidth;
}

/*
 int Tag::minheight(int width)
 { 
    int minheight;
 
    switch (type)
     {
     case TAG_TABLE: // minheight is the sum of rows minheights (children)
       {
	 unsigned int i;
	minheight = 0; // spaces between rows
    	for (i = 0; i< this->tags.size(); i++)
	   minheight += this->tags[i]->minheight(width);
 	break;
       }
     case TAG_TR: // minheight maximum of columns minwidths (children)
       {
	 unsigned int i;
	minheight = 0;
 	for (i = 0; i<this->tags.size(); i++)
	   {
	     if  (this->tags[i]->minheight(width) > minheight)
	       minheight = this->tags[i]->minheight(width);
 	  }
	 break;
       }
     case TAG_TD: // meisure it as a content length or sum of children lengths divided by width
     case TAG_P:
     case TAG_SPAN:
     case TAG_B:
     case TAG_I:
     case TAG_U:
       {
	 unsigned int i;
 	minheight = 0;
	 for (i = 0; i<tags.size(); i++)
	   if (tags[i]->isblock()) 
	     minheight++;
 
	 minheight += length()/width;
 	if ((1.0*length())/width > 0.0f)
	   minheight++;
	break;
       }
     case TAG_BR:
     case TAG_HR:
       minheight = 1;
       break;
     case TAG_UNKNOWN:
     default:
       minheight = 0;
     }
  
   return minheight;
 }*/

int Tag::length(void)
{
  unsigned int i;
  int length=0;
  
  if (content != NULL)
    length += strlen(content);

  for (i=0; i<tags.size(); i++)
    {
      length += tags[i]->length();
    }

  return length;
}

int Tag::isblock(void)
{
  return ((type==TAG_TABLE) || (type==TAG_TR) || (type==TAG_BR) || 
	  (type==TAG_HR) || (type==TAG_UL) || (type==TAG_OL) ||
	  (type==TAG_H1) || (type==TAG_H2) || (type==TAG_H3) ||
	  (type==TAG_IMG) || (type==TAG_LI));
}

/*
 current - space, to what data shoud be printed. width, height - total size of space.
 wx, wy  - starting point of window, we're currently printing in
 wwidth, wheight - end point of that window
 xoffset, yoffset - point, where are we in this window
 layer - which layers shoud be displayed
*/
char*** Tag::getformatted(char*** current, int width, int height, int wx, int wy, int wwidth, int wheight, int *xoffset, int *yoffset, int layer)
{
  char ***out = NULL;
  char **text = NULL;
  char **effects = NULL;
  int i, j;

  //  fprintf(stderr, "getformatted %d out[%d][%d] win(%d,%d;%d,%d) offs(%d,%d)\n", gettype(), width, height, wx, wy, wx+wwidth, wy+wheight, *xoffset, *yoffset);

  if (current == NULL)
    {
      out = (char***) malloc(2*sizeof(char**));
      for (i = 0; i<2; i++)
	{
	  out[i] = (char**) malloc(height*sizeof(char*));
	  for (j = 0; j<height; j++)
	    {
	      out[i][j] = (char*) malloc(width*sizeof(char)+1);
	      memset(out[i][j], ' ', width);
	      out[i][j][width] = 0;
	    }
	}
    }
  else
    out = current;

  text = out[0];
  effects = out[1];

  wx += style.getmarginleft();
  wy += style.getmargintop();
  wwidth -= style.getmarginright() + style.getmarginright();
  wheight -= style.getmarginbottom();

  if (wx<0 || wy<0 || wwidth>width || wheight>height)  // if we're excessing screen, there's no sense to print anything
    return out;

  switch (type)
    {
    case TAG_TABLE: // print all elements (TR tags) one under another
      {
	unsigned int a, i, rows=0;
	int part_wy, part_height;
	for (a=0; a<tags.size(); a++)
	  if (tags[a]->gettype() == TAG_TR)
	    rows++;

	if (rows>0)
	    part_height = wheight/rows;

	i=0;

	//fprintf(stderr, "have %d rows in %d wheight, give them height %d\n", rows, wheight, part_height);

	for (a=0; a<rows; a++)
	  {
	    while ((i<tags.size()) && (tags[i]->gettype() != TAG_TR)) 
	      i++;

	    part_wy = wy + a*part_height;
	    *yoffset = part_wy;
	    tags[i]->getformatted(out, width, height, wx, part_wy, wwidth, part_height, xoffset, yoffset, layer);
	    i++;
	  }
	break;
      }
    case TAG_TR: // print all elements (TD tags) one next to another
      {
	unsigned int a, i, cols=0;
	int part_wx, part_width;
	for (a=0; a<tags.size(); a++)
	  if (tags[a]->gettype() == TAG_TD)
	    cols++;

	if (cols>0)
	  part_width = wwidth/cols - 1;

	i=0;
	for (a=0; a<cols; a++)
	  {
	    while ((i<tags.size()) && (tags[i]->gettype() != TAG_TD)) 
	      i++;

	    part_wx = wx + a*(part_width + 1);
	    *yoffset = wy;
	    *xoffset = part_wx;
	    tags[i]->getformatted(out, width, height, part_wx, wy, part_width, wheight, xoffset, yoffset, layer);
	    i++;
	  }
	break;
      }
    case TAG_IMG:
      {
	int x, y;
	*xoffset = wx;
	*yoffset = wy;

	for (y=0; y<5; y++)
	  {
	    for (x=0; x<5; x++) // an image is 5x5 box of X chars
	      {
		text[*yoffset][*xoffset] = 'X';
		(*xoffset)++;
	      }
	    (*yoffset)++;
	  }

	break;
      }

    case TAG_SLIDE: // block tags, which set x and y to wx, wy
    case TAG_TD:
    case TAG_P:
    case TAG_H1:
    case TAG_H2:
    case TAG_H3:
    case TAG_UL:
    case TAG_OL:      
      *xoffset = wx;
      *yoffset = wy;
    case TAG_LI:
    case TAG_SPAN:
    case TAG_B:
      // set some special param to B and follow SPAN
    case TAG_I:
      // set some special param to I and follow SPAN
    case TAG_U:
      // set some special param to U and follow SPAN
      {
	unsigned int a;
	int maxchars;
	if (content!=NULL)
	  {
	    //fprintf(stderr, "contentlen=%d xofs=%d maxx=%d yofs=%d maxy=%d\n", strlen(content), *xoffset, wx+wwidth, *yoffset, wy+wheight);
	    maxchars = strnrchr(content, wx+wwidth - *xoffset+1, ' ');
	    maxchars = (maxchars == -1 || (signed)strlen(content)<wwidth) ? strlen(content) : maxchars;

	    switch (style.getalign())
	      {
	      case (ha_right): *xoffset += wwidth-maxchars; break;
	      case (ha_center): *xoffset += (wwidth-maxchars)/2; break;
	      case (ha_justify): break;
	      case (ha_left): break;
	      } // end of align case

	    for (a=0; a<strlen(content) && ((*xoffset)<wx+wwidth) && ((*yoffset)<=wy+wheight); a++)
	      {
		text[*yoffset][*xoffset] = content[a];
		(*xoffset)++;
		maxchars--;

		if ((*xoffset == wx+wwidth || maxchars==0) && a<strlen(content)-1)
		  {
		    (*yoffset)++;
		    *xoffset = wx;

		    while (a+1<strlen(content) && blank(content[a+1])) // skip the spaces, that we're on
		      a++;

		    maxchars = strnrchr(&content[a], wx+wwidth - *xoffset+1, ' ');
		    maxchars = (maxchars == -1 || (signed)strlen(&content[a])<wwidth) ? strlen(&content[a]) : maxchars;
		    
		    switch (style.getalign())
		      {
		      case (ha_right): *xoffset += wwidth-maxchars; break;
		      case (ha_center): *xoffset += (wwidth-maxchars)/2; break;
		      case (ha_justify): break;
		      case (ha_left): break;
		      } // end of align case
		  }


	      } // end of for a=0, ... loop
	  }
	else
	  for (a=0; a<tags.size(); a++)
	    {
	      if ((a>1) && tags[a-1]->isblock()) 
		{
		  wy = (*yoffset);
		  wheight = (wy+wheight<height) ? wheight : height - wy - 1;
		}

	      /*
	      if (a>0 && !tags[a-1]->isblock() && !tags[a]->isblock() && (*xoffset>wx)) // put space between two text tags
		{
		  char ca = tags[a-1]->getlastchar();
		  char cb = tags[a]->getfirstchar();
		  //if (!blank(ca) && !blank(cb))
		    //		    (*xoffset)++;
		}
	      */
	      
	      tags[a]->getformatted(out, width, height, wx, wy, wwidth, wheight, xoffset, yoffset, layer);
	    }

	break;
      }
    case TAG_BR:
      {
	(*yoffset)++;
	*xoffset=wx;
	break;
      }
    case TAG_HR:
      {
	if (*xoffset > wx)
	  {
	    (*yoffset)++;
	    *xoffset=wx;
	  }

	while (*xoffset < wx+wwidth)
	  {
	    text[*yoffset][*xoffset]='-';
	    (*xoffset)++;
	  }
	(*yoffset)++;
	*xoffset=wx;
	break;
      }
    case TAG_UNKNOWN:
    default:      
      break;
    }

  *xoffset += style.getmarginright();
  *yoffset += style.getmarginbottom();

  //  fprintf(stderr, "leaving %d out[%d][%d] win(%d,%d;%d,%d) offs(%d,%d)\n", gettype(), width, height, wx, wy, wx+wwidth, wy+wheight, *xoffset, *yoffset);
  return out;
}

char*** Tag::getformatted(int width, int height, int layer)
{
  int xoffset=0, yoffset=0;
  return getformatted(NULL, width, height, 0, 0, width, height, &xoffset, &yoffset, layer);
}
