/****
   LTPS - Linux Text Presentation System
                                       ****/

#include <stdio.h>
#include <ncurses.h>
#include <time.h>

#include "transition.h"

using namespace std;

Transition::Transition()
{
}

Transition::Transition(char* opts)
{
}

int Transition::execute(char*** data, int width, int height)
{
  int y;
  char** text = NULL, **effects = NULL;

  text = data[0];
  effects = data[1];
  
  move(0, 0);

  for (y=0; y<height; y++)
    mvprintw(y, 0, "%s", text[y]);

  refresh();

  return 0;
}

int Trans_Up2Down::execute(char*** data, int width, int height)
{
  int y=0;
  struct timespec sleeptime;
  char** text = NULL, **effects = NULL;

  text = data[0];
  effects = data[1];
  
  move(0, 0);

  for (y=0; y<height; y++)
    {
      sleeptime.tv_sec = 0;
      sleeptime.tv_nsec = 10000000;

      nanosleep(&sleeptime, NULL);
      mvprintw(y, 0, "%s", text[y]);
      refresh();
    }

  return 0;
}

int Trans_Left2Right::execute(char*** data, int width, int height)
{
  int x=0, y=0;
  struct timespec sleeptime;
  char** text = NULL, **effects = NULL;

  text = data[0];
  effects = data[1];
  
  move(0, 0);

  for (x=0; x<width; x++)
    {
      sleeptime.tv_sec = 0;
      sleeptime.tv_nsec = 10000000;

      nanosleep(&sleeptime, NULL);
      
      for (y=0; y<height; y++)
	{
	  mvaddch(y, x, text[y][x]);
	}
      refresh();
    }

  return 0;
}

int Trans_Center2Borders::execute(char*** data, int width, int height)
{
  int x0, y0, r, rmax, d;
  struct timespec sleeptime;
  char** text = NULL, **effects = NULL;

  text = data[0];
  effects = data[1];
  
  x0 = width/2;
  y0 = height/2;
  rmax = x0>y0 ? x0 : y0;
  
  for (r=0; r<rmax+1; r++)
    {
      for (d=y0-r; d<y0+r; d++)
	{
	  if ((0 <= d) && (d < height) && (x0+r < width))
	      mvaddch(d, x0+r, text[d][x0+r]);
	  if ((0 <= d) && (d < height) && (0 <= x0-r))
	      mvaddch(d, x0-r, text[d][x0-r]);

	}
      for (d=x0-r; d<=x0+r; d++)
	{
	  if ((0 <= d) && (d < width) && (y0+r < height))
	      mvaddch(y0+r, d, text[y0+r][d]);
	  if ((0 <= d) && (d < width) && (0 <= y0-r))
	      mvaddch(y0-r, d, text[y0-r][d]);
	}

      refresh();

      sleeptime.tv_sec = 0;
      sleeptime.tv_nsec = 10000000;
      
      nanosleep(&sleeptime, NULL);
    }

  return 0;
}

int Trans_MoveFromLeft::execute(char*** data, int width, int height)
{
  int x, movx, y;
  struct timespec sleeptime;
  char** text = NULL, **effects = NULL;

  text = data[0];
  effects = data[1];
  
  for (movx=1; movx<=width; movx++)
    {      
      for (x=0; x<movx; x++)
	{
	  for (y=0; y<height; y++)
	    {
	      mvaddch(y, x, text[y][width-movx+x]);
	    }
	}
      refresh();

      sleeptime.tv_sec = 0;
      sleeptime.tv_nsec = 10000000;

      nanosleep(&sleeptime, NULL);
    }

  return 0;
}

int Trans_Matrix::execute(char*** data, int width, int height)
{
  int x=0, y=0;
  struct timespec sleeptime;
  char** text = NULL, **effects = NULL;

  text = data[0];
  effects = data[1];
  
  for (x=0; x<width; x++)
    {
      sleeptime.tv_sec = 1;

      nanosleep(&sleeptime, NULL);
      
      for (y=0; y<height; y++)
	{
	  mvaddch(y, x, text[y][x]);
	}
      refresh();
    }

  return 0;
}
